# -*- coding: utf-8 -*-

# Songwrite 2
# Copyright (C) 2007 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later verson.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import division
import os, os.path, sys, math, bisect, locale, codecs
from cStringIO import StringIO

import gobject, gtk, cairo, pango, pangocairo
if gtk.pygtk_version < (2, 7, 0): import cairo.gtk

from editobj2.introsp  import *
from editobj2.observe  import *
from editobj2.undoredo import *
import editobj2.editor_gtk as editor_gtk

import songwrite2.globdef as globdef
import songwrite2.model   as model
import songwrite2.stemml  as stemml
import songwrite2.player  as player
import songwrite2.__editobj2__

zoom_levels = [0.25, 0.35, 0.5, 0.75, 1.0, 1.5, 2.0, 3.0, 4.0]


SELECTION_COLOR  = None
SELECTION_COLOR2 = None

CLIPBOARD        = None
CLIPBOARD_SOURCE = None
CLIPBOARD_X1     = 0
CLIPBOARD_Y1     = 0
CLIPBOARD_X2     = 0
CLIPBOARD_Y2     = 0

CLIPBOARD2_NAME = "_SONGWRITE2_NOTES"
CLIPBOARD2 = gtk.Clipboard(selection = CLIPBOARD2_NAME)

def find_view_children(o):
  if isinstance(o, model.View) and hasattr(o, "strings"):
    return o.strings
  return []

class BaseCanvas(object):
  def __init__(self, song, zoom = 1.0):
    self.x                   = 0
    self.y                   = 0
    self.width               = 0
    self.height              = 0
    self.song                = song
    self.start_x             = editor_gtk.SMALL_ICON_SIZE + 10
    self.zoom                = zoom
    self.selection_x1        = 1000000
    self.selection_y1        = 1000000
    self.selection_x2        = -1
    self.selection_y2        = -1
    self.default_font_size   = 16.0
    self.default_line_height = 0
    self.cursor              = None
    self.cursor_drawer       = None
    self.canvas              = None
    self.drawers             = None
    self.partition_2_drawer  = {}
    
  def reset(self):
    for drawer in self.drawers: drawer.destroy()
    self.drawers = []
    
  def note_string_id(self, note):
    return self.partition_2_drawer[note.partition].note_string_id(note)
  
  def y_2_drawer(self, y):
    for drawer in self.drawers:
      if y < drawer.y + drawer.height: return drawer
    return None
    
  def x_2_time(self, x):
    return (x - self.start_x + self.x) / self.zoom
  
  def time_2_x(self, time):
    return self.start_x - self.x + time * self.zoom
  

class Canvas(gtk.Fixed, BaseCanvas):
  def __init__(self, main, song, scrolled, zoom = 1.0):
    BaseCanvas.__init__(self, song, zoom)
    gtk.Fixed.__init__(self)
    self.set_has_window(1)
    
    global SELECTION_COLOR
    if SELECTION_COLOR is None:
      global SELECTION_COLOR2
      color = gtk.Entry().rc_get_style().bg[gtk.STATE_SELECTED]
      SELECTION_COLOR2 = color.red / 65535.0, color.green / 65535.0, color.blue / 65535.0
      SELECTION_COLOR = [1.0 - 0.35 * (1.0 - x) for x in SELECTION_COLOR2]
      color = gtk.Entry().rc_get_style().bg[gtk.STATE_PRELIGHT]
      
    self.canvas = gtk.DrawingArea()
    self.put(self.canvas, 0, 0)
    
    #self.width             = 0
    #self.height            = 0
    self.main              = main
    self.song              = song
    self.hadjustment       = scrolled.get_hadjustment()
    self.vadjustment       = scrolled.get_vadjustment()
    #self.x                 = 0
    #self.y                 = 0
    #self.zoom              = 1.0
    self.selections        = set()
    self.last_selections   = set()
    self.drag_selections   = set()
    self.click_x           = 0
    self.click_y           = 0
    self.drag_start_x      = 0
    self.drag_start_y      = 0
    #self.selection_x1      = 1000000
    #self.selection_y1      = 1000000
    #self.selection_x2      = -1
    #self.selection_y2      = -1
    self.last_selection_x1 = 1000000
    self.last_selection_y1 = 1000000
    self.last_selection_x2 = -1
    self.last_selection_y2 = -1
    self.need_update_selection = 0
    self.current_duration  = 96
    self.previous_text     = "" # Previously typed text
    self.edit_timeout      = 0
    self.first_time        = 1
    self.x_before_tracking = -1
    
    self.default_font_size   = self.get_style().font_desc.get_size() / pango.SCALE * pangocairo.cairo_font_map_get_default().get_resolution() / 72.0
    self.default_line_height = 0
    #self.drawers             = None
    #self.partition_2_drawer  = {}
    
    self.canvas.add_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.KEY_PRESS_MASK | gtk.gdk.BUTTON1_MOTION_MASK)
    self.canvas.set_flags(gtk.CAN_FOCUS)
    self.canvas.connect('expose_event'             , self.on_expose)
    self.canvas.connect('key_press_event'          , self.on_key_press)
    self.canvas.connect('motion_notify_event'      , self.on_mouse_motion)
    self.canvas.connect('button_press_event'       , self.on_button_press)
    self.canvas.connect('button_release_event'     , self.on_button_release)
    self.connect('drag-drop'                , self.on_drag_drop)
    self.connect('drag-data-received'       , self.on_drag_data_received)
    self.connect('drag-data-get'            , self.on_drag_data_get)
    self.connect('drag-data-delete'         , self.on_drag_data_delete)
    self.connect('size-allocate'            , self.on_resized)
    self.connect('scroll-event'            , self.on_before_scroll)
    self.vadjustment.connect("value-changed", self.on_scroll)
    self.hadjustment.connect("value-changed", self.on_scroll)
    self.vadjustment.step_increment = self.hadjustment.step_increment = 10
    
    observe(song                        , self.song_listener)
    observe(song.playlist               , self.playlist_listener)
    observe(song.playlist.playlist_items, self.playlist_listener)
    observe(song.mesures                , self.mesures_listener)
    for mesure in song.mesures:
      observe(mesure, self.mesure_listener)
      
    observe(song.partitions, self.partitions_listener)
    for partition in song.partitions:
      observe(partition      , self.partition_listener)
      if isinstance(partition, model.Partition):
        observe_tree(partition.view , self.view_listener, find_view_children)
        #if hasattr(partition.view, "strings"): observe(partition.view.strings, self.strings_listener)
        observe(partition.notes, self.notes_listener)
        for note in partition.notes:
          observe(note, self.note_listener)
          
    self.drag_source_set(gtk.gdk.BUTTON1_MASK, [(CLIPBOARD2_NAME, 0, 0)], gtk.gdk.ACTION_MOVE | gtk.gdk.ACTION_COPY)
    self.drag_source_targets = self.drag_source_get_target_list()
    self.drag_source_unset()
    
    self.drag_dest_set(gtk.DEST_DEFAULT_HIGHLIGHT | gtk.DEST_DEFAULT_DROP, [(CLIPBOARD2_NAME, 0, 0)], gtk.gdk.ACTION_MOVE | gtk.gdk.ACTION_COPY)
    self.connect("drag-motion", self.on_drag_motion)
    
  def on_drag_motion(self, widget, drag, x, y, timestamp):
    if drag.actions & gtk.gdk.ACTION_MOVE:
      drag.drag_status(gtk.gdk.ACTION_MOVE, timestamp)
    else:
      drag.drag_status(gtk.gdk.ACTION_COPY, timestamp)
    return 1
  
  def play_tracker(self, time):
    if time == -1:
      self.render_selection()
      self.selection_x1 = self.last_selection_x1
      self.selection_y1 = self.last_selection_y1
      self.selection_x2 = self.last_selection_x2
      self.selection_y2 = self.last_selection_y2
      self.render_selection()
      self.hadjustment.value = max(0, self.x_before_tracking)
      self.x_before_tracking = -1
      
    else:
      if self.x_before_tracking == -1: self.x_before_tracking = max(0, self.x)
      self.render_selection()
      self.selection_x1 = self.time_2_x(time) + self.x
      self.selection_y1 = self.drawers[1].y + self.y
      self.selection_x2 = self.selection_x1 + 2 * self.char_h_size
      self.selection_y2 = self.drawers[-1].y + self.drawers[-1].height + self.y
      self.render_selection()

      if   self.selection_x2 > self.x + self.width * 0.75:
        self.hadjustment.value = min(self.hadjustment.upper, self.selection_x2 - self.width * 0.1)
      elif self.selection_x1 < self.x:
        self.hadjustment.value = max(self.hadjustment.lower, self.selection_x2 - self.width * 0.1)
        
  def destroy(self):
    for drawer in self.drawers: drawer.destroy()
    unobserve(self.cursor)
    unobserve(self.song                        , self.song_listener)
    unobserve(self.song.playlist               , self.playlist_listener)
    unobserve(self.song.playlist.playlist_items, self.playlist_listener)
    unobserve(self.song.mesures                , self.mesures_listener)
    unobserve(self.song.partitions             , self.partitions_listener)
    
    for mesure in self.song.mesures:
      unobserve(mesure, self.mesure_listener)
      
    for partition in self.song.partitions:
      unobserve(partition, self.partition_listener)
      if isinstance(partition, model.Partition):
        unobserve_tree(partition.view, self.view_listener, find_view_children)
        unobserve(partition.notes, self.notes_listener)
        for note in partition.notes:
          unobserve(note, self.note_listener)
          
    gtk.Fixed.destroy(self)
    
  def update_time(self):
    for drawer in self.drawers:
      if isinstance(drawer, LyricsDrawer): drawer.update_melody()
        
  def get_selected_time(self):
    if self.selections     : return list(self.selections     )[0].time
    if self.last_selections: return list(self.last_selections)[0].time
    return 0
  
  def get_selected_mesures(self):
    return list(set([self.song.mesure_at(note.time) for note in self.selections]))
  
  def get_selected_partitions(self):
    return list(set([note.partition for note in self.selections]))
  
  def song_listener(self, song, type, new, old):
    self.main.set_title(u"Songwrite2 -- %s" % song.title)
    self.render_all()
    
  def playlist_listener(self, playlist, type, new, old):
    self.render_all()
  
  def mesures_listener(self, obj, type, new, old):
    new = set(new)
    old = set(old)
    for mesure in new - old: observe  (mesure, self.mesure_listener)
    for mesure in old - new: unobserve(mesure, self.mesure_listener)
    
  def mesure_listener(self, mesure, type, new, old):
    self.render_all()
    
  def partitions_listener(self, obj, type, new, old):
    new = set(new)
    old = set(old)
    for partition in new - old:
      observe  (partition, self.partition_listener)
      if isinstance(partition, model.Partition):
        observe_tree(partition.view , self.view_listener, find_view_children)
        #if hasattr(partition.view, "strings"): observe(partition.view.strings, self.strings_listener)
        observe(partition.notes, self.notes_listener)
        for note in partition.notes: observe(note, self.note_listener)
        
    for partition in old - new:
      unobserve(partition, self.partition_listener)
      if isinstance(partition, model.Partition):
        unobserve_tree(partition.view , self.view_listener, find_view_children)
        #if hasattr(partition.view, "strings"): unobserve(partition.view.strings, self.strings_listener)
        unobserve(partition.notes, self.notes_listener)
        for note in partition.notes: unobserve(note, self.note_listener)
        
    self.deselect_all()
    self.reset()
    self.render_all()
    
  def partition_listener(self, partition, type, new, old):
    if type is object:
      if   (new.get("view") != old.get("view")):
        unobserve_tree(old.get("view"), self.view_listener, find_view_children)
        observe_tree  (new.get("view"), self.view_listener, find_view_children)
        
        self.deselect_all()
        self.reset()
        self.render_all()
      if   (new.get("g8") != old.get("g8")) or (new.get("tonality") != old.get("tonality")):
        self.deselect_all()
        self.reset()
        self.render_all()
      elif (new.get("instrument") != old.get("instrument")) or (new.get("header") != old.get("header")) or (new.get("visible") != old.get("visible")):
        self.render_all()
        
    self.partition_2_drawer[partition].partition_listener(partition, type, new, old)
    
  def view_listener(self, view, type, new, old):
    self.render_all()
    
  def strings_listener(self, view, type, new, old):
    self.render_all()
    
  def notes_listener(self, obj, type, new, old):
    new = set(new)
    old = set(old)
    for note in new - old: observe  (note, self.note_listener)
    for note in old - new: unobserve(note, self.note_listener)
    
  def note_listener(self, note, type, new, old):
    if note in self.selections:
      self.need_update_selection = 1
      if (note is self.cursor) and (note.value > 0):
        self.cursor = None
      self.render_selection()
      
    self.current_duration = note.duration
    self.partition_2_drawer[note.partition].render_note(note)
    
  def set_cursor(self, cursor, cursor_drawer):
    self.cursor        = cursor
    self.cursor_drawer = cursor_drawer
    
  def set_zoom(self, zoom):
    self.selection_x1 = (self.selection_x1 - self.start_x) * zoom / self.zoom + self.start_x
    self.selection_x2 = (self.selection_x2 - self.start_x) * zoom / self.zoom + self.start_x
    self.hadjustment.value = self.hadjustment.value  / self.zoom * zoom
    self.zoom = zoom
    self.render_pixel(0, 0, self.width, self.height)
    self.update_time()
    self.main.zoom_menu.get_child().set_text(u"%s%%" % int(zoom * 100.0))
   
  def delayed_edit(self, note):
    if self.edit_timeout: gobject.source_remove(self.edit_timeout)
    self.edit_timeout = gobject.timeout_add(40, self.edit, note)
    
  def edit(self, note):
    if getattr(self.main, "notebook", None):
      mesures = self.get_selected_mesures()
      if len(mesures) == 1: mesure = mesures[0]
      else:                 mesure = ObjectPack(mesures)

      partitions = self.get_selected_partitions()
      if partitions:
        if len(partitions) == 1:
          partition = partitions[0]
        else:
          partition = ObjectPack(partitions)
          partition.song = ObjectPack([p.song for p in partitions])
        self.main.notebook.edit(-3, partition)

      self.main.notebook.edit(-2, mesure)
      self.main.notebook.edit(-1, note)

  def arrange_selection_at_fret(self, fret):
    old_string_ids = dict([(note, note.string_id) for note in self.selections if hasattr(note, "string_id")])
    def do_it():
      for note, string_id in old_string_ids.iteritems():
        if isinstance(note.partition.view, model.TablatureView):
          strings = note.partition.view.strings
          if note.value - strings[string_id].base_note < fret:
            while (string_id < len(strings) - 1) and (note.value - strings[string_id].base_note < fret):
              string_id += 1
          else:
            while (string_id > 0) and (note.value - strings[string_id - 1].base_note >= fret):
              string_id -= 1
          note.string_id = string_id
      self.render_all()
      
    def undo_it():
      for note, string_id in old_string_ids.iteritems():
        note.string_id = string_id
      self.render_all()
    UndoableOperation(do_it, undo_it, _(u"arrange at fret #%s" % fret), self.main.undo_stack)
    
    
  def select_all(self):
    self.deselect_all()
    for partition in self.song.partitions:
      if isinstance(partition, model.Partition):
        drawer = self.partition_2_drawer[partition]
        for note in partition.notes:
          self.add_selection(note, render = 0, edit = 1, *drawer.note_dimensions(note))
    self.render_selection()
    
  def deselect_all(self):
    if self.cursor: self.cursor_drawer.destroy_cursor()
    if self.selection_x2 != -1:
      self.render_selection()
      self.selection_x1     = 1000000
      self.selection_y1     = 1000000
      self.selection_x2     = -1
      self.selection_y2     = -1
    self.selections      = set()
    self.cursor          = None
    self.previous_text   = ""
    
  def extend_selection(self, x1, y1, x2, y2):
    if self.selection_x1 > x1:
      self.selection_x1 = x1
      if x1 < self.x + self.start_x: self.hadjustment.value = x1 - self.start_x
    
    if self.selection_y1 > y1:
      self.selection_y1 = y1
      if y1 < self.y: self.vadjustment.value = y1
      
    if self.selection_x2 < x2:
      self.selection_x2 = x2
      if x2 > self.x + self.width: self.hadjustment.value = x2 - self.width
        
    if self.selection_y2 < y2:
      self.selection_y2 = y2
      if y2 > self.y + self.height: self.vadjustment.value = y2 - self.height
      
  def add_selection(self, note, x1, y1, x2, y2, render = 1, edit = 1):
    self.selections.add(note)
    self.extend_selection(x1 + self.x, y1 + self.y, x2 + self.x, y2 + self.y)
    if not note is self.cursor:
      global CLIPBOARD, CLIPBOARD_SOURCE, CLIPBOARD_X1, CLIPBOARD_Y1, CLIPBOARD_X2, CLIPBOARD_Y2
      CLIPBOARD_SOURCE = self
      CLIPBOARD        = self.last_selections   = self.selections
      CLIPBOARD_X1     = self.last_selection_x1 = self.selection_x1
      CLIPBOARD_Y1     = self.last_selection_y1 = self.selection_y1
      CLIPBOARD_X2     = self.last_selection_x2 = self.selection_x2
      CLIPBOARD_Y2     = self.last_selection_y2 = self.selection_y2
      
      CLIPBOARD2.set_with_data([(CLIPBOARD2_NAME, 0, 0)], self.clipboard_get, self.clipboard_clear, None)
    if render: self.render_selection()
    if edit  : self.delayed_edit(note)
    
  def clipboard_clear(self, clipboard, userdata): pass
  
  def clipboard_get(self, clipboard, selection_data, info, userdata):
    _xml    = StringIO()
    xml     = codecs.lookup("utf8")[3](_xml)
    context = model._XMLContext()
    xml.write("""<?xml version="1.0" encoding="utf-8"?>\n<notes click_x="0">""")
    
    min_y = 10000.0
    for note in CLIPBOARD:
      drawer = self.partition_2_drawer[note.partition]
      x1, y1, x2, y2 = drawer.note_dimensions(note)
      y = (y1 + y2) // 2
      if y < min_y: min_y = y
      
    for note in CLIPBOARD:
      drawer = self.partition_2_drawer[note.partition]
      x1, y1, x2, y2 = drawer.note_dimensions(note)
      nb_char = len(drawer.note_text(note))
      x = x1 + (self.char_h_size * nb_char) // 2 - CLIPBOARD_X1
      y = (y1 + y2) // 2 - min_y
      note.__xml__(xml, context, ' x="%r" y="%r" string_pitch="%s" view_type="%s"' % (x, y, drawer.strings[drawer.note_string_id(note)].base_note, note.partition.view.__class__.__name__[:-4].lower()))
    xml.write("</notes>")
    selection_data.set(selection_data.target, 8, xml.getvalue())
    
  def on_button_press(self, obj, event):
    self.canvas.grab_focus()
    #if event.button == 1:
    self.click_x = event.x + self.x
    self.click_y = event.y + self.y
    if   event.button == 1:
      if self.selections and (self.selection_x1 <= self.click_x <= self.selection_x2) and (self.selection_y1 <= self.click_y <= self.selection_y2):
        pass
      else:
        self.deselect_all()
        if event.x > self.start_x:
          for drawer in self.drawers:
            if drawer.y <= event.y <= drawer.y + drawer.height: drawer.on_button_press(event)
            
    elif event.button == 2:
#       def callback(clipboard, selection_data, user_data = None):
#         xml = selection_data.data
#         print xml
#         orig_time, notes = stemml.parse(StringIO(xml))
#         song = model.Song()
#         for note
#         canvas = Canvas(None, None, song)
        
#       self.clipboard.request_contents("songwrite notes", callback)
      
      if CLIPBOARD:
        drag = CLIPBOARD_SOURCE.drag_begin(self.drag_source_targets, gtk.gdk.ACTION_COPY, 2, event)
        CLIPBOARD_SOURCE.drag_start_x    = self.drag_start_x = CLIPBOARD_X1 + 20
        CLIPBOARD_SOURCE.drag_start_y    = self.drag_start_y = CLIPBOARD_Y1 + 20
        CLIPBOARD_SOURCE.drag_selections = set(CLIPBOARD)
        
        width  = int(CLIPBOARD_X2 - CLIPBOARD_X1) + 2
        height = int(CLIPBOARD_Y2 - CLIPBOARD_Y1) + 2
        pixmap = gtk.gdk.Pixmap(self.window, width, height)
        ctx = pixmap.cairo_create()
        
        CLIPBOARD_SOURCE.draw_drag_icon(ctx, CLIPBOARD_X1 - 1, CLIPBOARD_Y1 - 1, width, height)
        pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, 1, 8, max(width, 100), height) # It seems that GTK doesn't like small drag icons !
        pixbuf.fill(0x00000000)
        pixbuf.get_from_drawable(pixmap, self.window.get_colormap(), 0, 0, 0, 0, width, height)
        pixbuf = pixbuf.add_alpha(1, chr(255), chr(255), chr(255))
        drag.set_icon_pixbuf(pixbuf, 20, 20)
        
      #else:
      #  self.deselect_all()
      #  if event.x > self.start_x:
      #    for drawer in self.drawers:
      #      if drawer.y <= event.y <= drawer.y + drawer.height: drawer.on_button_press(event)
      #  self.on_note_paste()
        
    elif event.button == 3:
      self.deselect_all()
      if event.x > self.start_x:
        for drawer in self.drawers:
          if drawer.y <= event.y <= drawer.y + drawer.height: drawer.on_button_press(event)
          
      notes      = list(self.selections)
      if notes:
        old_values = [note.value for note in notes]
        
        drawer = self.partition_2_drawer[notes[0].partition]
        player.play_note(notes[0].partition.instrument,
                         abs(notes[0].value))
        
        def do_it(notes = notes):
          for note in notes:
            drawer = self.partition_2_drawer[note.partition]
            if note.value < 0: note.partition.add_note(note)
            note.value = abs(note.value)
        def undo_it(notes = notes, old_values = old_values):
          for i in range(len(notes)):
            notes[i].value = old_values[i]
            if notes[i].value < 0: notes[i].partition.remove(notes[i])

        UndoableOperation(do_it, undo_it, _(u"note change"), self.main.undo_stack)
      
  def on_button_release(self, obj, event):
    if (not self.selections) and (self.selection_x2 != -1):
      time1 = self.x_2_time(self.selection_x1 - self.x)
      time2 = self.x_2_time(self.selection_x2 - self.x)
      y1 = self.selection_y1 - self.y
      y2 = self.selection_y2 - self.y
      
      self.deselect_all()
      for partition in self.song.partitions:
        if not isinstance(partition, model.Partition): continue
        drawer = self.partition_2_drawer[partition]
        string_id1 = drawer.y_2_string_id(y1 + drawer.string_height // 2, 1)
        string_id2 = drawer.y_2_string_id(y2 - drawer.string_height // 2, 1)
        
        for note in partition.notes_at(time1, time2):
          if string_id1 <= drawer.note_string_id(note) <= string_id2:
            self.add_selection(note, render = 0, edit = 0, *drawer.note_dimensions(note))
            
      if self.selections:
        self.render_selection()
        if len(self.selections) == 1:
          self.delayed_edit(tuple(self.selections)[0])
        else:
          pack = ObjectPack(list(self.selections))
          self.delayed_edit(pack)
#    elif self.selections and (self.selection_x1 <= self.click_x <= self.selection_x2) and (self.selection_y1 <= self.click_y <= self.selection_y2):
#      self.deselect_all()
#      if event.x > self.start_x:
#        for drawer in self.drawers:
#          if drawer.y <= event.y <= drawer.y + drawer.height: drawer.on_button_press(event)
          
  def on_mouse_motion(self, obj, event):
    if self.drag_check_threshold(int(self.click_x - self.x), int(self.click_y - self.y), int(event.x), int(event.y)):
      if self.selections and (self.selection_x1 <= self.click_x <= self.selection_x2) and (self.selection_y1 <= self.click_y <= self.selection_y2) and not self.cursor:
        drag = self.drag_begin(self.drag_source_targets, gtk.gdk.ACTION_MOVE | gtk.gdk.ACTION_COPY, 1, event)
        
        self.drag_start_x    = self.click_x
        self.drag_start_y    = self.click_y
        self.drag_selections = set(self.last_selections)
        
        width  = int(self.selection_x2 - self.selection_x1) + 2
        height = int(self.selection_y2 - self.selection_y1) + 2
        pixmap = gtk.gdk.Pixmap(self.window, width, height)
        ctx = pixmap.cairo_create()
        
        self.draw_drag_icon(ctx, self.selection_x1 - 1, self.selection_y1 - 1, width, height)
        pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, 1, 8, max(width, 100), height) # It seems that GTK doesn't like small drag icons !
        pixbuf.fill(0x00000000)
        pixbuf.get_from_drawable(pixmap, self.window.get_colormap(), 0, 0, 0, 0, width, height)
        pixbuf = pixbuf.add_alpha(1, chr(255), chr(255), chr(255))
        drag.set_icon_pixbuf(pixbuf, int(self.click_x - self.selection_x1), int(self.click_y - self.selection_y1))
        
      else:
        if self.selections: self.deselect_all()
        old_sel_x1 = self.selection_x1
        old_sel_y1 = self.selection_y1
        old_sel_x2 = self.selection_x2
        old_sel_y2 = self.selection_y2
        if   event.x < self.start_x: self.hadjustment.value = max(0, self.hadjustment.value - 10)
        elif event.x > self.width  : self.hadjustment.value = min(max(self.hadjustment.upper - self.width, 0), self.hadjustment.value + 10)
        if   event.y < 0           : self.vadjustment.value = max(0, self.vadjustment.value - 10)
        elif event.y > self.height : self.vadjustment.value = min(max(self.vadjustment.upper - self.height, 0), self.vadjustment.value + 10)
        self.selection_x1 = max(min(self.click_x, event.x + self.x), self.start_x)
        self.selection_y1 = min(self.click_y, event.y + self.y)
        self.selection_x2 = max(self.click_x, event.x + self.x)
        self.selection_y2 = max(self.click_y, event.y + self.y)
        self.render(min(old_sel_x1, self.selection_x1) - 1,
                    min(old_sel_y1, self.selection_y1) - 1,
                    max(old_sel_x2, self.selection_x2) - min(old_sel_x1, self.selection_x1) + 2,
                    max(old_sel_y2, self.selection_y2) - min(old_sel_y1, self.selection_y1) + 2)
        
  def on_drag_drop(self, widget, drag_context, x, y, timestamp): return True
  
  def on_drag_data_get(self, widget, drag_context, selection_data, info, timestamp):
    _xml    = StringIO()
    xml     = codecs.lookup("utf8")[3](_xml)
    context = model._XMLContext()
    #xml.write("""<?xml version="1.0" encoding="utf-8"?>\n<notes click_time="%s">""" % self.x_2_time(self.drag_start_x))
    xml.write("""<?xml version="1.0" encoding="utf-8"?>\n<notes click_x="%s">""" % (self.drag_start_x - self.selection_x1))
    
    for note in self.drag_selections:
      drawer = self.partition_2_drawer[note.partition]
      x1, y1, x2, y2 = drawer.note_dimensions(note)
      nb_char = len(drawer.note_text(note))
      x = x1 + (self.char_h_size * nb_char) // 2 - self.drag_start_x
      y = (y1 + y2) // 2 - self.drag_start_y + self.y
      note.__xml__(xml, context, ' x="%r" y="%r" string_pitch="%s" view_type="%s"' % (x, y, drawer.strings[drawer.note_string_id(note)].base_note, note.partition.view.__class__.__name__[:-4].lower()))
    xml.write("</notes>")
    selection_data.set(selection_data.target, 8, xml.getvalue())
    
  def on_drag_data_received(self, widget, drag_context, x0, y0, selection_data, info, timestamp):
    xml = selection_data.data
    
    #print xml
    from cStringIO import StringIO
    click_x, notes = stemml.parse(StringIO(xml))
    
    orig_time = click_x / self.zoom
    dest_time = self.x_2_time(x0)
    
    self.paste_notes(notes, dest_time, orig_time, y0)
    
  def on_note_paste(self):
    xml = CLIPBOARD2.wait_for_contents(CLIPBOARD2_NAME).data
    
    from cStringIO import StringIO
    click_x, notes = stemml.parse(StringIO(xml))
    
    if not self.selections: return
    note      = tuple(self.selections)[0]
    drawer    = self.partition_2_drawer[note.partition]
    y0        = drawer.string_id_2_y(drawer.note_string_id(note))
    dest_time = note.time
    orig_time = 0
    
    self.paste_notes(notes, dest_time, orig_time, y0)
    
  def paste_notes(self, notes, dest_time, orig_time, y0):
    orig_time += min([note.time for note in notes])
      
    dt = int(0.4 + float(abs(dest_time - orig_time)) / self.current_duration) * self.current_duration
    if dest_time < orig_time: dt = -dt
    
    notes_data     = []
    previous_notes = {}
    for note in notes:
      y = y0 + note.y
      drawer = self.y_2_drawer(y)
      if not isinstance(drawer, PartitionDrawer): continue
      time = note.time + dt
      if time < 0: continue
      if   isinstance(drawer, TablatureDrawer):
        if (note.view_type == u"tablature") or (note.view_type == u"drums"):
          string_id = drawer.y_2_string_id(y)
          if string_id is None: continue
        else:
          string_id = drawer.default_string_id(note.value)
      elif isinstance(drawer, DrumsDrawer):
        if (note.view_type == u"tablature") or (note.view_type == u"drums"):
          string_id = drawer.y_2_string_id(y)
          if string_id is None: continue
      else: string_id = 0
      notes_data.append((note, drawer, time, string_id))
      for previous_note in drawer.partition.notes_at(time, time):
        if string_id == drawer.note_string_id(previous_note):
          drawer_previous_notes = previous_notes.get(drawer)
          if not drawer_previous_notes: drawer_previous_notes = previous_notes[drawer] = []
          drawer_previous_notes.append(previous_note)
          break
        
    def do_it(notes_data = notes_data):
      for drawer, notes in previous_notes.items():
        drawer.partition.remove(*notes)
        
      new_notes = {}
      for note, drawer, time, string_id in notes_data:
        new_note = model.Note(drawer.partition, time, note.duration, note.value, note.volume)
        new_note.fx = note.fx
        if note.bend_pitch: new_note.bend_pitch = note.bend_pitch
        if   isinstance(drawer, TablatureDrawer):
          if   (note.view_type == "tablature"):
            new_note.string_id = string_id
            new_note.value     = note.value - note.string_pitch + drawer.strings[string_id].base_note
          else:
            new_note.string_id = string_id
            #new_note.string_id = drawer.default_string_id(note.value)
        elif isinstance(drawer, DrumsDrawer):
          new_note.string_id = string_id
          new_note.value     = drawer.strings[string_id].base_note
          
        drawer_new_notes = new_notes.get(drawer)
        if not drawer_new_notes: drawer_new_notes = new_notes[drawer] = []
        drawer_new_notes.append(new_note)
        
      self.deselect_all()
      
      for drawer, notes in new_notes.items():
        drawer.partition.add_note(*notes)
        
        for note in notes:
          self.add_selection(note, render = 0, edit = 0, *drawer.note_dimensions(note))
          
      self.render_all()

    def undo_it(notes = notes):
      new_notes = {}
      for note, drawer, time, string_id in notes_data:
        drawer_new_notes = new_notes.get(drawer)
        if not drawer_new_notes: drawer_new_notes = new_notes[drawer] = []
        for new_note in drawer.partition.notes_at(time, time):
          if isinstance(drawer, TablatureDrawer) or isinstance(drawer, DrumsDrawer):
            if string_id == drawer.note_string_id(new_note):
              drawer_new_notes.append(new_note)
              break
          else:
            if new_note.value == note.value:
              drawer_new_notes.append(new_note)
              break
            
      for drawer, notes in new_notes.items():
        drawer.partition.remove(*notes)
        
      for drawer, notes in previous_notes.items():
        drawer.partition.add_note(*notes)
        
      self.render_all()
      
    UndoableOperation(do_it, undo_it, _(u"add %s note(s)") % len(notes), self.main.undo_stack)



  def round_time_to_current_duration(self, time, delta = 0):
    duration = self.current_duration
    
    if duration ==  96:
      if (self.song.mesure_at(time) or self.song.mesures[-1]).rythm2 == 8: duration /= 2
    if duration == 144:
      if (self.song.mesure_at(time) or self.song.mesures[-1]).rythm2 == 4: duration /= 3
    else:
      if duration / 1.5 in model.DURATIONS.keys(): duration /= 3
      
    return max(0, int((delta + (time // duration)) * duration))
    
  def on_drag_data_delete(self, widget, drag_context):
    # Some notes int the selection may have been already deleted, if the selection was pasted partly
    # over the selection itself.
    notes = [note for note in self.drag_selections if note.partition and note in note.partition.notes]
    self.delete_notes(notes)
    
  def on_key_press(self, obj, event):
    #print dir(event), event.keyval
    keyval = event.keyval
    if   keyval == 65361: # left
      if self.selections:
        sel = tuple(self.selections)[0]
        time = sel.time
        string_id = self.note_string_id(sel)
        note = sel.partition.note_before_pred(sel, lambda a: self.note_string_id(a) == string_id)
        time0 = self.round_time_to_current_duration(time, -1)
        #time0 = ((time - 1) // self.current_duration) * self.current_duration
        if note and note.time >= time0:
          self.deselect_all()
          self.partition_2_drawer[sel.partition].select_note(note)
        elif time > 0:
          self.deselect_all()
          self.partition_2_drawer[sel.partition].select_at(time0, string_id)
          
    elif keyval == 65363: # right
      if self.selections:
        sel = tuple(self.selections)[0]
        time = sel.time
        string_id = self.note_string_id(sel)
        note = sel.partition.note_after_pred(sel, lambda a: self.note_string_id(a) == string_id)
        time0 = self.round_time_to_current_duration(time, 1)
        #if note and note.time <= time + self.current_duration:
        if note and note.time <= time0:
          self.deselect_all()
          self.partition_2_drawer[sel.partition].select_note(note)
        else:
          self.deselect_all()
          #time += self.current_duration
          self.partition_2_drawer[sel.partition].select_at(time0, string_id)
          
    elif keyval == 65362: # up
      if self.selections:
        sel = tuple(self.selections)[0]
        string_id = self.note_string_id(sel)
        if string_id > 0:
          self.deselect_all()
          self.partition_2_drawer[sel.partition].select_at(sel.time, string_id - 1)
        else:
          i = self.drawers.index(self.partition_2_drawer[sel.partition])
          if i > 0:
            drawer = self.drawers[i - 1]
            if isinstance(drawer, PartitionDrawer):
              self.deselect_all()
              drawer.select_at(sel.time, len(drawer.strings) - 1)
              
    elif keyval == 65364: # down
      if self.selections:
        sel = tuple(self.selections)[0]
        string_id = self.note_string_id(sel)
        if string_id < len(self.partition_2_drawer[sel.partition].strings) - 1:
          self.deselect_all()
          self.partition_2_drawer[sel.partition].select_at(sel.time, string_id + 1)
        else:
          i = self.drawers.index(self.partition_2_drawer[sel.partition])
          if i < len(self.drawers) - 1:
            drawer = self.drawers[i + 1]
            if isinstance(drawer, PartitionDrawer):
              self.deselect_all()
              drawer.select_at(sel.time, 0)
              
    elif keyval == 65360: # Orig
      if self.selections:
        note   = list(self.selections)[0]
        drawer = self.partition_2_drawer[note.partition]
        string_id = drawer.note_string_id(note)
        self.deselect_all()
        drawer.select_at(0, string_id)
        
    elif keyval == 65367: # End
      if self.selections:
        note   = list(self.selections)[0]
        drawer = self.partition_2_drawer[note.partition]
        string_id = drawer.note_string_id(note)
        self.deselect_all()
        drawer.select_at(self.song.mesures[-1].end_time() - self.current_duration, string_id)
        
    elif (keyval == 47) or (keyval == 42): # / or *
      if keyval == 47: notes = [note for note in self.selections if note.base_duration() > 12 ]
      else:            notes = [note for note in self.selections if note.base_duration() < 384]
      def decrease_duration(notes = notes):
        for note in notes: self.current_duration = note.duration = note.duration // 2
      def increase_duration(notes = notes):
        for note in notes: self.current_duration = note.duration = note.duration *  2
        
      if keyval == 47: UndoableOperation(decrease_duration, increase_duration, _(u"decrease note duration"), self.main.undo_stack)
      else:            UndoableOperation(increase_duration, decrease_duration, _(u"increase note duration"), self.main.undo_stack)
      
    elif keyval == 46: # .
      self.toggle_dotted()
      
    elif (keyval == 43) or (keyval == 65453) or (keyval == 61): # + ou =
      self.shift_selections_values(1)
      
    elif (keyval == 45) or (keyval == 65455): # -
      self.shift_selections_values(-1)
      
    elif (48 <= keyval <= 57) or (65456 <= keyval <= 65465): # Numbers
      if keyval <= 57: nb = keyval - 48
      else:            nb = keyval - 65456
      self.previous_text += str(nb)
      notes      = list(self.selections)
      if notes:
        text       = self.previous_text
        old_values = [note.value for note in notes]
        
        drawer = self.partition_2_drawer[notes[0].partition]
        player.play_note(notes[0].partition.instrument,
                         drawer.strings[drawer.note_string_id(notes[0])].text_2_value(text))
        
        def do_it(notes = notes, text = text):
          for note in notes:
            drawer = self.partition_2_drawer[note.partition]
            if note.value < 0: note.partition.add_note(note)
            note.value = drawer.strings[drawer.note_string_id(note)].text_2_value(text)
        def undo_it(notes = notes, old_values = old_values):
          for i in range(len(notes)):
            notes[i].value = old_values[i]
            if notes[i].value < 0: notes[i].partition.remove(notes[i])

        UndoableOperation(do_it, undo_it, _(u"note change"), self.main.undo_stack)
      
    elif (keyval == 65535) or (keyval == 65288): # Del
      if self.selections: self.delete_notes(self.selections)
      
    elif (keyval == 104) or (keyval == 108) or (keyval == 112): # h, p, l
      if self.selections: self.set_selections_fx(u"link")
      
    elif keyval == 100: # d
      if self.selections: self.set_selections_fx(u"dead")
      
    elif keyval == 116: # t
      if self.selections: self.set_selections_fx(u"tremolo")
      
    elif keyval == 98: # b
      if self.selections: self.set_selections_fx(u"bend")
      
    elif keyval == 115: # s
      if self.selections: self.set_selections_fx(u"slide")
      
    elif keyval == 114: # r
      if self.selections: self.set_selections_fx(u"roll")
      
    elif keyval == 110: # n
      if self.selections: self.set_selections_fx(u"")
      
    elif (keyval == 65293) or (keyval == 65421): # return
      if self.selections:
        notes = list(self.selections)
        def stress(notes = notes):
          for note in notes: note.volume = 255
        def unstress(notes = notes):
          for note in notes: note.volume = 204
        if notes[0].volume == 255: UndoableOperation(unstress, stress, _(u"unstress note"), self.main.undo_stack)
        else:                      UndoableOperation(stress, unstress, _(u"stress note"), self.main.undo_stack)
      
    elif keyval == 113: # q
      self.main.on_close()
      
    elif keyval == 32: # Space
      if player.is_playing(): player.stop()
      else: self.main.on_play_from_here()
      
    elif (keyval == 118) and (event.state & gtk.gdk.CONTROL_MASK): # C-V
      self.on_note_paste()
      
    elif (keyval == 97) and (event.state & gtk.gdk.CONTROL_MASK): # C-A
      self.select_all()
    return 1
  
  def shift_selections_values(self, d):
      notes      = list(self.selections)
      old_values = [note.value for note in notes]
      def do_it(notes = notes):
        for note in notes:
          drawer = self.partition_2_drawer[note.partition]
          if note.value < 0: note.partition.add_note(note)

#           if 0 and isinstance(drawer, StaffDrawer):
#             value_0 = note.value
#             if d > 0: sens =  1
#             else:     sens = -1
#             while 1:
#               print note.value, value_0, drawer.note_string_id(note), drawer.get_string(value_0).id, d
#               note.value = abs(note.value) + sens
#               string = drawer.get_string(note.value, 1)
#               if string.base_note != note.value: continue
#               if string.id == drawer.get_string(value_0).id - d: break
              
          note.value = abs(note.value) + d
            
          if isinstance(drawer, TablatureDrawer) and (note.value < drawer.strings[drawer.note_string_id(note)].base_note):
            note.value = drawer.strings[drawer.note_string_id(note)].base_note
            
      def undo_it(notes = notes, old_values = old_values):
        for i in range(len(notes)):
          notes[i].value = old_values[i]
          if notes[i].value < 0: notes[i].partition.remove(notes[i])
          
      if d > 0: UndoableOperation(do_it, undo_it, _(u"increase notes pitch"), self.main.undo_stack)
      else:     UndoableOperation(do_it, undo_it, _(u"decrease notes pitch"), self.main.undo_stack)

  def set_selections_volume(self, volume):
    if self.selections:
      notes      = list(self.selections)
      old_values = [note.volume for note in notes]
      def do_it(notes = notes):
        for note in notes: note.volume = volume
      def undo_it(notes = notes):
        for i in range(len(notes)): notes[i].volume = old_values[i]
      UndoableOperation(do_it, undo_it, _(u"change of %s") % _(u"volume"), self.main.undo_stack)
      
  def toggle_dotted(self, *args):
    notes      = list(self.selections)
    old_values = [note.duration for note in notes]
    def do_it(notes = notes):
      dotted = not notes[0].is_dotted()
      for note in notes:
        if note.is_dotted() != dotted:
          if dotted: self.current_duration = note.duration = int(note.duration * 1.5)
          else:      self.current_duration = note.duration = int(note.duration / 1.5)
    def undo_it(notes = notes, old_values = old_values):
      for i in range(len(notes)): notes[i].duration = old_values[i]

    UndoableOperation(do_it, undo_it, _(u"dotted note"), self.main.undo_stack)
    
  def toggle_triplet(self, *args):
    notes      = list(self.selections)
    old_values = [note.duration for note in notes]
    def do_it(notes = notes):
      triplet = not notes[0].is_triplet()
      for note in notes:
        if note.is_triplet() != triplet:
          if triplet: self.current_duration = note.duration = int(note.duration * 2 / 3)
          else:       self.current_duration = note.duration = int(note.duration * 3 / 2)
    def undo_it(notes = notes, old_values = old_values):
      for i in range(len(notes)): notes[i].duration = old_values[i]
      
    UndoableOperation(do_it, undo_it, _(u"triplet note"), self.main.undo_stack)
    
  def delete_notes(self, notes):
    if not notes: return
    notes = list(notes)
    if notes == [self.cursor]: return
    
    def do_it(notes = notes):
      sel = notes[0]
      drawer = self.partition_2_drawer[sel.partition]
      partition_2_notes = {}
      for note in notes:
        l = partition_2_notes.get(note.partition)
        if not l: partition_2_notes[note.partition] = [note]
        else: l.append(note)
        drawer.render_note(note)
      for partition, notes in partition_2_notes.items(): partition.remove(*notes)
      if set(notes) & self.selections:
        self.deselect_all()
        drawer.select_at(sel.time, drawer.note_string_id(sel), 0)
    def undo_it(notes = notes):
      self.deselect_all()
      partition_2_notes = {}
      for note in notes:
        l = partition_2_notes.get(note.partition)
        if not l: partition_2_notes[note.partition] = [note]
        else: l.append(note)
      for partition, notes in partition_2_notes.items():
        partition.add_note(*notes)
        drawer = self.partition_2_drawer[partition]
        for note in notes:
          self.add_selection(note, render = 0, edit = 0, *drawer.note_dimensions(note))
          drawer.render_note(note)

    UndoableOperation(do_it, undo_it, _(u"delete %s note(s)") % len(notes), self.main.undo_stack)

  def set_selections_fx(self, fx, bend_pitch = None):
    notes      = list(self.selections)
    old_values = [note.fx for note in notes]
    if not bend_pitch is None: old_bend_pitches = [note.bend_pitch for note in notes]
    def do_it():
      for note in notes:
        note.set_fx(fx)
        if not bend_pitch is None: note.bend_pitch = bend_pitch
    def undo_it():
      for i in range(len(notes)):
        notes[i].set_fx(old_values[i])
        if not bend_pitch is None:
          if old_bend_pitches[i] == 0.0: del notes[i].bend_pitch
          else:                          notes[i].bend_pitch = old_bend_pitches[i]
    UndoableOperation(do_it, undo_it, _(u"set special effect %s") % fx, self.main.undo_stack)
    
  def render_pixel(self, x, y, width, height):
    self.canvas.queue_draw_area(int(x), int(y), int(width), int(height))
    
  def render_all(self):
    self.canvas.queue_draw_area(0, 0, self.width, self.height)
    
  def render(self, x, y, width, height):
    self.canvas.queue_draw_area(int(x - self.x), int(y - self.y), int(width), int(height))
    
  def render_selection(self):
    self.canvas.queue_draw_area(int(self.selection_x1 - self.x) - 1, int(self.selection_y1 - self.y) - 1, int(self.selection_x2 - self.selection_x1) + 3, int(self.selection_y2 - self.selection_y1) + 3)
    
  def on_resized(self, obj, allocation):
    x, y, self.width, self.height = self.allocation
    self.canvas.set_size_request(self.width, self.height)
    
  def on_expose(self, obj, event):
    self.draw(event.area.x, event.area.y, event.area.width, event.area.height)
    
  def draw(self, x, y, width, height):
    self.x = int(self.hadjustment.value)
    self.y = int(self.vadjustment.value)
    
    if self.need_update_selection:
      self.selection_x1     = 1000000
      self.selection_y1     = 1000000
      self.selection_x2     = -1
      self.selection_y2     = -1
      for note in self.selections: self.add_selection(note, render = 0, edit = 0, *self.partition_2_drawer[note.partition].note_dimensions(note))
      self.need_update_selection = 0
      
    if gtk.pygtk_version < (2,7,0): ctx = cairo.gtk.gdk_cairo_create(self.canvas.window)
    else:                           ctx = self.canvas.window.cairo_create()
    left, top, self.width, self.height = self.allocation
    
    ctx.set_source_rgb(1.0, 1.0, 1.0)
    ctx.rectangle(x, y, width, height)
    ctx.fill()
    
    first_drawer = None
    if not self.drawers: # first time
      ctx.select_font_face('sans', cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL)
      ctx.set_font_size(self.default_font_size)
      self.default_ascent, self.default_descent, self.default_line_height, max_x_advance, max_y_advance = ctx.font_extents()
      self.char_h_size = self.default_line_height // 2 # XXX
      
      self.drawers = [SongDrawer(self, self.song)]
      for partition in self.song.partitions:
        if   isinstance(partition     , model.Lyrics       ): drawer = LyricsDrawer   (self, partition)
        elif isinstance(partition.view, model.TablatureView): drawer = TablatureDrawer(self, partition)
        elif isinstance(partition.view, model.StaffView    ): drawer = StaffDrawer    (self, partition)
        elif isinstance(partition.view, model.DrumsView    ): drawer = DrumsDrawer    (self, partition)
        else:                                                 drawer = partition.view.get_drawer(self)
        if drawer:
          self.drawers.append(drawer)
          if (not first_drawer) and self.first_time:
            first_drawer = drawer
            self.first_time = 0
            
    if self.selection_x2 != -1:
      ctx.save()
      ctx.rectangle(self.start_x - 1, 0, self.width, self.height)
      ctx.clip()
      ctx.set_line_width(2.0)
      draw_round_rect(ctx, self.selection_x1 - self.x, self.selection_y1 - self.y, self.selection_x2 - self.x, self.selection_y2 - self.y, 8)
      ctx.set_source_rgb(*SELECTION_COLOR)
      ctx.fill_preserve()
      if self.canvas.is_focus():
        ctx.set_source_rgb(*SELECTION_COLOR2)
      ctx.stroke()
      ctx.set_line_width(1.0)
      ctx.restore()
      
    dy = -self.y
    total_height = 0
    for drawer in self.drawers:
      drawer.y = dy
      drawer.draw(ctx, x, y, width, height)
      dy           += drawer.height + 10
      total_height += drawer.height + 10
      
    self.hadjustment.upper          = (self.song.mesures[-1].end_time() + self.song.mesures[-1].duration) * self.zoom
    self.hadjustment.page_increment = self.width // 2
    self.hadjustment.page_size      = self.width
    self.vadjustment.upper          = total_height
    self.vadjustment.page_increment = self.height // 2
    self.vadjustment.page_size      = self.height
    
    if first_drawer: first_drawer.select_at(0, 0, 0)
    
    if (total_height <= self.height) and (self.y != 0):
      self.vadjustment.set_value(0)
      self.render_all()
      
  def draw_drag_icon(self, ctx, x, y, width, height):
    _x = self.x
    _y = self.y
    self.x = x
    self.y = y
    
    ctx.set_source_rgb(1.0, 1.0, 1.0)
    ctx.rectangle(0, 0, width, height)
    ctx.fill()
    
    ctx.set_line_width(2.0)
    draw_round_rect(ctx, 1, 1, width - 2, height - 2, 8)
    ctx.set_source_rgb(1.0, 1.0, 1.0)
    ctx.fill_preserve()
    ctx.set_source_rgb(*SELECTION_COLOR2)
    ctx.stroke()
    ctx.set_line_width(1.0)
    
    dy = -self.y
    for drawer in self.drawers:
      ctx.new_path()
      _drawer_y = drawer.y
      drawer.y = dy
      drawer.draw(ctx, 0, dy, width, height, 1)
      drawer.y = _drawer_y
      dy      += drawer.height + 10
      
    self.x = _x
    self.y = _y
    
  def change_zoom(self, incr):
    i = zoom_levels.index(self.zoom) + incr
    if   i < 0:                    i = 0
    elif i > len(zoom_levels) - 1: i = len(zoom_levels) - 1
    self.set_zoom(zoom_levels[i])
    
  def on_before_scroll(self, obj, event):
    if event.state & gtk.gdk.CONTROL_MASK:
      if   (event.direction == gtk.gdk.SCROLL_DOWN) or (event.direction == gtk.gdk.SCROLL_LEFT):
        self.change_zoom(-1)
      else:
        self.change_zoom( 1)
      return 1
    
  def on_scroll(self, obj):
    self.render_pixel(0, 0, self.width, self.height)
    
    
    
class Drawer(object):
  def __init__(self, canvas):
    self.canvas = canvas
    self.y      = 0
    self.height = 0
    
  def draw(self, ctx, x, y, width, height, drag = 0):
    return (y < self.y + self.height) and (self.y < y + height)
  
  def destroy(self): pass
  
class SongDrawer(Drawer):
  def __init__(self, canvas, song):
    Drawer.__init__(self, canvas)
    self.song = song
    
  def on_button_press(self, event):
    pass
  
  def draw(self, ctx, x, y, width, height, drag = 0):
    if self.song.authors: title = "%s (%s)" % (self.song.title, self.song.authors)
    else:                 title = self.song.title
    ctx.select_font_face('sans', cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_BOLD)
    ctx.set_font_size(self.canvas.default_font_size * 1.2)
    ctx.set_source_rgb(0,0,0)
    x_bearing, y_bearing, width, height, x_advance, y_advance = ctx.text_extents(title)
    
    if y < height + height + y_bearing:
      ctx.move_to((self.canvas.width - width) / 2, height + self.y)
      ctx.show_text(title)
    ctx.set_font_size(self.canvas.default_font_size)
    ctx.select_font_face('sans', cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL)
    
    if self.song.comments:
      self.height = draw_multiline_text(ctx, self.song.comments, 10, height + 10 + self.y, self.canvas.width - 20) + 10 - self.y
    else:
      self.height = height + 10
      
    return 1
  
  
class PartitionDrawer(Drawer):
  show_header = 1
  def __init__(self, canvas, partition):
    Drawer.__init__(self, canvas)
    self.partition = partition
    self.start_y   = 15
    self.stem_extra_height = 0
    self.stem_extra_height_bottom = 0
    self.up_stem_ratio = 0.5
    self.canvas.partition_2_drawer[partition] = self

  def partition_listener(self, partition, type, new, old): pass
  
  def y_2_string_id(self, y, force = 0):
    string_id = int((y - self.y - self.stem_extra_height - self.start_y) // self.string_height)
    if not force:
      if   string_id < 0:                     return None
      elif string_id > len(self.strings) - 1: return None
    return string_id
  
  def string_id_2_y(self, string_id):
    return self.y + self.start_y + self.stem_extra_height + self.string_height * (string_id + 0.5)
  
  def on_button_press(self, event):
    string_id = self.y_2_string_id(event.y)
    if not string_id is None:
      time = self.canvas.x_2_time(event.x)
      self.select_at(time, string_id)
      
  def select_at(self, time, string_id, fix_time = 1):
    #if fix_time: time0 = time - time % self.canvas.current_duration
    if fix_time: time0 = self.canvas.round_time_to_current_duration(time)
    else:        time0 = time
    notes = self.partition.notes_at(time0, time)
    notes = [note for note in notes if self.note_string_id(note) == string_id]
    for note in notes:
      if note.end_time() > time:
        self.select_note(note)
        break
    else:
      for note in notes:
        if note.time == time0:
          self.select_note(note)
          break
      else:
        cursor          = model.Note(self.partition)
        cursor.time     = time0
        cursor.duration = self.canvas.current_duration
        if isinstance(self, TablatureDrawer):
          cursor.string_id = string_id
        if isinstance(self, StaffDrawer):
          cursor.value = -(self.strings[string_id].base_note + self.strings[string_id].alteration)
        else:
          cursor.value = -self.strings[string_id].base_note
        while cursor.time >= self.canvas.song.mesures[-1].end_time():
          mesure = self.canvas.song.add_mesure()
          self.canvas.render(self.canvas.time_2_x(mesure.time) - 1, 0, mesure.duration * self.canvas.zoom + 2, self.canvas.height)
        observe(cursor, self.canvas.note_listener)
        self.canvas.set_cursor(cursor, self)
        self.select_note(cursor)
        self.render_note(cursor)
        
  def destroy(self):
    self.destroy_cursor(render = 0)
    Drawer.destroy(self)
    
  def destroy_cursor(self, render = 1):
    cursor = self.canvas.cursor
    unobserve(cursor)
    self.canvas.cursor        = None
    self.canvas.cursor_drawer = None
    if render: self.render_note(cursor)
    
  def select_note(self, note):
    self.canvas.add_selection(note, *self.note_dimensions(note))
    
  def render_note(self, note):
    mesure = self.canvas.song.mesure_at(note)
    if mesure.rythm2 == 8: time1 = (note.link_start() // 144) * 144; time2 = max(note.end_time(), time1 + 144)
    else:                  time1 = (note.link_start() //  96) *  96; time2 = max(note.end_time(), time1 +  96)
    self.canvas.render_pixel(
      self.canvas.time_2_x(time1),
      #self.y + self.start_y - 1,
      self.y,
      (time2 - time1) * self.canvas.zoom + 2,
      #self.height - self.start_y + 3,
      self.height + self.canvas.default_line_height // 2 + self.canvas.default_ascent // 2,
      )
    
  def note_dimensions(self, note):
    x = self.canvas.time_2_x(note.time)
    y = self.string_id_2_y(self.note_string_id(note)) - self.string_height // 2
    return (
      x,
      y,
      x + note.duration * self.canvas.zoom,
      y + self.string_height,
      )
  
  def draw_mesure(self, ctx, time, mesure, y, height):
    x       = self.canvas.time_2_x(time)
    ny      = y
    nheight = height
    symbols = self.partition.song.playlist.symbols.get(mesure)
    if symbols:
      bold = 1
      for symbol in symbols:
        if symbol.startswith(r"\repeat"):
          ctx.arc(x + 5, y + 0.30 * height, 1.5, 0, 2.0 * math.pi)
          ctx.stroke_preserve()
          ctx.set_source_rgb(0.8, 0.0, 0.0)
          ctx.fill()
          ctx.set_source_rgb(0.0, 0.0, 0.0)
          
          ctx.arc(x + 5, y + 0.70 * height, 1.5, 0, 2.0 * math.pi)
          ctx.stroke_preserve()
          ctx.set_source_rgb(0.8, 0.0, 0.0)
          ctx.fill()
          ctx.set_source_rgb(0.0, 0.0, 0.0)
        elif (symbol == "} % end repeat") or (symbol == "} % end alternative"):
          ctx.arc(x - 5, y + 0.30 * height, 1.5, 0, 2.0 * math.pi)
          ctx.stroke_preserve()
          ctx.set_source_rgb(0.8, 0.0, 0.0)
          ctx.fill()
          ctx.set_source_rgb(0.0, 0.0, 0.0)
          
          ctx.arc(x - 5, y + 0.70 * height, 1.5, 0, 2.0 * math.pi)
          ctx.stroke_preserve()
          ctx.set_source_rgb(0.8, 0.0, 0.0)
          ctx.fill()
          ctx.set_source_rgb(0.0, 0.0, 0.0)
        elif symbol.startswith(r"{ % start alternative"):
          ny      = self.y
          nheight = height + y - ny
          alt = symbol[22:]
          ctx.move_to(x + 5, ny + self.canvas.default_line_height * 0.85)
          ctx.show_text(alt)
          ctx.move_to(x, ny + 0.5)
          ctx.rel_line_to(60, 0)
          ctx.stroke()
    else:
      bold = (mesure is None) or (time == 0)
      
    if bold:
      ctx.set_line_width(2)
      ctx.move_to    (x, ny)
      ctx.rel_line_to(0, nheight)
      ctx.stroke()
      ctx.set_line_width(1)
    else:
      ctx.set_line_width(1)
      ctx.move_to    (x + 0.5, ny)
      ctx.rel_line_to(0      , nheight)
      ctx.stroke()
      
  def draw_strings(self, ctx, x, y, width, height):
    string_x  = x
    string_x2 = x + width
    
    if (string_x < string_x2):
      i = 0
      for string in self.strings:
        string_y = self.string_id_2_y(i)
        i += 1
        if y - 2 < string_y < y + height + 2:
          w = string.width()
          if   w == -1: continue
          elif w == 0:
            ctx.set_line_width(2.0)
            ctx.set_source_rgb(0.4, 0.4, 0.4)
            ctx.move_to(string_x , string_y + .4)
            ctx.line_to(string_x2, string_y + .4)
            ctx.stroke()
          elif w == 1:
            ctx.set_line_width(1.75)
            ctx.set_source_rgb(0.45, 0.45, 0.45)
            ctx.move_to(string_x , string_y + .4)
            ctx.line_to(string_x2, string_y + .4)
            ctx.stroke()
          elif w == 2:
            ctx.set_line_width(1.7)
            ctx.set_source_rgb(0.5, 0.5, 0.5)
            ctx.move_to(string_x , string_y + .4)
            ctx.line_to(string_x2, string_y + .4)
            ctx.stroke()
          elif w == 3:
            ctx.set_line_width(1.6)
            ctx.set_source_rgb(0.55, 0.55, 0.55)
            ctx.move_to(string_x , string_y + .4)
            ctx.line_to(string_x2, string_y + .4)
            ctx.stroke()
          elif w == 4:
            ctx.set_line_width(1.5)
            ctx.set_source_rgb(0.6, 0.6, 0.6)
            ctx.move_to(string_x , string_y + .4)
            ctx.line_to(string_x2, string_y + .4)
            ctx.stroke()
          elif w == 5:
            ctx.set_line_width(1.2)
            ctx.set_source_rgb(0.75, 0.75, 0.75)
            ctx.move_to(string_x , string_y + .2)
            ctx.line_to(string_x2, string_y + .2)
            ctx.stroke()
          elif w == 6:
            ctx.set_line_width(1.1)
            ctx.set_source_rgb(0.8, 0.8, 0.8)
            ctx.move_to(string_x , string_y + .2)
            ctx.line_to(string_x2, string_y + .2)
            ctx.stroke()
          elif w == 7:
            ctx.set_line_width(1.0)
            ctx.set_source_rgb(0.85, 0.85, 0.85)
            ctx.move_to(string_x , string_y + .3)
            ctx.line_to(string_x2, string_y + .3)
            ctx.stroke()
          elif w == 8:
            ctx.set_line_width(1.2)
            ctx.set_dash([3, 3], string_x % 6)
            ctx.set_source_rgb(0.75, 0.75, 0.75)
            ctx.move_to(string_x , string_y + .2)
            ctx.line_to(string_x2, string_y + .2)
            ctx.stroke()
            ctx.set_dash([])
          elif w == 9:
            ctx.set_line_width(0.6)
            ctx.set_source_rgb(0.0, 0.0, 0.0)
            ctx.move_to(string_x , string_y + .3)
            ctx.line_to(string_x2, string_y + .3)
            ctx.stroke()
      ctx.set_line_width(1.0)
    self.height = int(self.start_y + self.stem_extra_height + self.string_height * len(self.strings) + self.stem_extra_height_bottom + 1)
    
  def draw(self, ctx, x, y, width, height, drag = 0, draw_icon = 1):
    if x < self.canvas.start_x:
      width = width + x - self.canvas.start_x
      x = self.canvas.start_x
    if not drag:
      if self.show_header:
        self.start_y = 15 + draw_multiline_text(ctx, self.partition.header, 0, self.y + 2, self.canvas.width - 20) - self.y
      else:
        self.start_y = 15
#     ctx.set_source_rgb(0.9, 1.0, 0.9)
#     ctx.rectangle(x, self.y, width, self.height)
#     ctx.fill()
    
#     ctx.set_source_rgb(0.8, 1.0, 0.8)
#     ctx.rectangle(x, self.y, width, self.start_y + self.stem_extra_height)
#     ctx.fill()
    
#     ctx.set_source_rgb(0.7, 1.0, 0.7)
#     ctx.rectangle(x, self.y, width, self.start_y)
#     ctx.fill()
    
#     ctx.set_source_rgb(0.0, 0.0, 0.0)
    
    if Drawer.draw(self, ctx, x, y, width, height, drag):
      ctx.save()
      if not drag:
        self.draw_strings(ctx, x, y, width, height)
        
        ctx.rectangle(self.canvas.start_x - 1, 0, self.canvas.width, self.canvas.height)
        ctx.clip()
        
        bar_y      = self.y + self.start_y + self.stem_extra_height + self.string_height * 0.5
        bar_height = self.string_height * (len(self.strings) - 1)
        ctx.set_source_rgb(0, 0, 0)
        for mesure in self.partition.song.mesures:
          #if self.canvas.time_2_x(mesure.time) >= x:
          if x - 60 <= self.canvas.time_2_x(mesure.time) <= x + width:
            self.draw_mesure(ctx, mesure.time, mesure, bar_y, bar_height)
        self.draw_mesure(ctx, mesure.end_time(), None, bar_y, bar_height)
        
      time1  = self.canvas.x_2_time(x - self.canvas.start_x)
      time2  = self.canvas.x_2_time(x + width)
      mesure = self.canvas.song.mesure_at(time2)
      if mesure:
        if mesure.rythm2 == 8:
          time1 = (    time1 // 144) * 144
          time2 = (1 + time2 // 144) * 144
        else:
          time1 = (    time1 //  96) *  96
          time2 = (1 + time2 //  96) *  96
          
      notes = self.partition.notes_at(time1, time2)
      if self.canvas.cursor and (self.canvas.cursor.partition is self.partition):
        if time1 <= self.canvas.cursor <= time2:
          bisect.insort(notes, self.canvas.cursor)
          
      current_time = -1
      previous     = None
      notes_y = {}
      ctx.set_font_size(self.canvas.default_font_size)
      ctx.select_font_face('sans', cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL)
      
#       for note in notes:
#         string_id = self.note_string_id(note)
#         note_y = notes_y[note] = self.string_id_2_y(string_id)
        
#         if current_time != note.time:
#           if (current_time != -1) and not drag:
#             self.draw_stem_and_beam_for_chord(ctx, notes_at_time, notes_y, previous, note)
#             previous = notes_at_time[0]
#           current_time  = note.time
#           notes_at_time = []
          
#         notes_at_time.append(note)
#         ctx.set_source_rgb(0.0, 0.0, 0.0)
#         self.draw_note(ctx, note, string_id, note_y)
        
#       if (current_time != -1) and not drag: self.draw_stem_and_beam_for_chord(ctx, notes_at_time, notes_y, previous, None)


      chords = []
      if drag:
        for note in notes:
          string_id = self.note_string_id(note)
          note_y = notes_y[note] = self.string_id_2_y(string_id)
          ctx.set_source_rgb(0.0, 0.0, 0.0)
          self.draw_note(ctx, note, string_id, note_y)
          
      else:
        for note in notes:
          string_id = self.note_string_id(note)
          note_y = notes_y[note] = self.string_id_2_y(string_id)
          
          if current_time != note.time:
            if (current_time != -1): chords.append(notes_at_time)
            current_time  = note.time
            notes_at_time = []
            
          notes_at_time.append(note)
          
          ctx.set_source_rgb(0.0, 0.0, 0.0)
          self.draw_note(ctx, note, string_id, note_y)
          
        if current_time != -1: chords.append(notes_at_time)
        
        for i in range(0, len(chords)):
          if   i == 0: previous = None
          elif len(chords[i - 1]) == 1: previous = chords[i - 1][0]
          else:
            min_duration = min([note.duration for note in chords[i - 1]])
            for previous in chords[i - 1]:
              if previous.duration == min_duration: break
          if   i == len(chords) - 1: next = None
          elif len(chords[i + 1]) == 1: next = chords[i + 1][0]
          else:
            min_duration = min([note.duration for note in chords[i + 1]])
            for next in chords[i + 1]:
              if next.duration == min_duration: break
          self.draw_stem_and_beam_for_chord(ctx, chords[i], notes_y, previous, next)

      
      ctx.restore()
      
      if (not drag) and draw_icon:
        # Draws the partition icon
        #pixbuf = editor_gtk.load_small_icon(description(self.partition.__class__).icon_filename_for(self.partition))
        #self.canvas.canvas.window.draw_pixbuf(None, pixbuf, 0, 0, 4, int(self.y + self.start_y // 2 + self.stem_extra_height - self.stem_extra_height_bottom + self.height // 2 - editor_gtk.SMALL_ICON_SIZE // 2), -1, -1)
        draw_bitmap(self.canvas, ctx,
                    description(self.partition.__class__).icon_filename_for(self.partition),
                    4,
                    int(self.y + self.start_y + self.stem_extra_height + (self.height - self.start_y - self.stem_extra_height - self.stem_extra_height_bottom) // 2 - editor_gtk.SMALL_ICON_SIZE // 2),
                    0.5)
        
  def draw_stem_and_beam_for_chord(self, ctx, notes, notes_y, previous, next):
    if len(notes) == 1:
      l = notes
    else:
      l = {}
      for note in notes: l[self.note_string_id(note)] = note
      l = l.items()
      l.sort()
      l = [note for (string_id, note) in l]
      
    x = self.canvas.time_2_x(l[0].time) + 0.5
    
    mesure = self.canvas.song.mesure_at(l[0].time)
    
    self.draw_stem_and_beam(ctx, x, l[ 0], self.y + self.start_y, notes_y[l[ 0]] - self.string_height // 2, mesure, previous, next)
    
    #if len(l) > 1:
    #  self.draw_stem_and_beam(ctx, x, l[ 0], self.y + self.start_y, notes_y[l[ 0]] - self.string_height // 2, mesure, previous, next)
    #  self.draw_stem_and_beam(ctx, x, l[-1], self.y + self.height , notes_y[l[-1]] + self.string_height // 2, mesure, previous, next)
    #else:
    #  if self.note_string_id(l[0]) < len(self.strings) * self.up_stem_ratio:
    #    self.draw_stem_and_beam(ctx, x, l[ 0], self.y + self.start_y, notes_y[l[ 0]] - self.string_height // 2, mesure, previous, next)
    #  else:
    #    self.draw_stem_and_beam(ctx, x, l[ 0], self.y + self.height , notes_y[l[ 0]] + self.string_height // 2, mesure, previous, next)
    
  def note_color(self, ctx, note):
    if note.volume == 255:
      ctx.set_source_rgb(0.85, 0.0, 0.0)
    else:
      intensity = 0.7 - min(0.7, 0.7 * note.volume / 204.0)
      ctx.set_source_rgb(intensity, intensity, intensity)
      
  def draw_note(self, ctx, note, string_id, y):
    self.note_color(ctx, note)
    x = self.canvas.time_2_x(note.time)
    ctx.move_to(x + 3, y + self.canvas.default_ascent / 2 - self.canvas.default_descent / 2)
    if note.is_dotted(): ctx.show_text(self.note_text(note) + ".")
    else:                ctx.show_text(self.note_text(note))
    if note.fx: getattr(self, "draw_note_fx_%s" % note.fx)(ctx, note, string_id)
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    
  def draw_note_fx_link(self, ctx, note, string_id):
    ax1, ay1, ax2, ay2 = self.note_dimensions(note)
    if note.linked_to: bx1, by1, bx2, by2 = self.note_dimensions(note.linked_to)
    else:              bx1, by1, bx2, by2 = ax1 + note.duration, ay1, ax2 + note.duration, ay2
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(ax1 + 3 + self.canvas.char_h_size // 2, ay2 - self.canvas.default_descent)
    ctx.curve_to(
      ax1 + 6 + self.canvas.char_h_size // 2, ay2 - self.canvas.default_descent + 8,
      bx1     + self.canvas.char_h_size // 2, by2 - self.canvas.default_descent + 8,
      bx1 + 3 + self.canvas.char_h_size // 2, by2 - self.canvas.default_descent,
      )
    ctx.stroke()
    
  def draw_note_fx_slide(self, ctx, note, string_id):
    ax1, ay1, ax2, ay2 = self.note_dimensions(note)
    if note.linked_to: bx1, by1, bx2, by2 = self.note_dimensions(note.linked_to)
    else:              bx1, by1, bx2, by2 = ax1 + note.duration, ay1, ax2 + note.duration, ay2
    bx1 -= 3
    anb_char = len(self.note_text(note))
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(ax1 + 5 + self.canvas.char_h_size * anb_char, (ay1 + ay2) // 2 + self.canvas.default_descent)
    ctx.line_to(bx1 + 2, (by1 + by2) // 2 + self.canvas.default_descent)
    ctx.rel_line_to(-5, -5)
    ctx.rel_move_to( 5,  5)
    ctx.rel_line_to(-5,  5)
    ctx.stroke()
    
  def draw_note_fx_bend(self, ctx, note, string_id):
    x1, y1, x2, y2 = self.note_dimensions(note)
    x2 -= 3
    anb_char = len(self.note_text(note))
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x1 + 5 + self.canvas.char_h_size * anb_char, (y1 + y2) // 2 + self.canvas.default_descent)
    dx = min(x2 - x1, self.canvas.char_h_size * 3)
    ctx.rel_curve_to(
       12,  0,
       dx, -self.string_height // 2 + 5,
       dx, -self.string_height // 2,
      )
    ctx.rel_line_to(-5,  1)
    ctx.rel_move_to( 5, -1)
    ctx.rel_line_to( 1,  5)
    ctx.stroke()
    ctx.set_font_size(self.canvas.default_font_size // 2)
    ctx.move_to(x1 + dx * 0.3 + anb_char * self.canvas.char_h_size, (y1 + y2) // 2)
    ctx.show_text(locale.str(note.bend_pitch))
    ctx.set_font_size(self.canvas.default_font_size)
    
  def draw_note_fx_tremolo(self, ctx, note, string_id):
    x1, y1, x2, ty2 = self.note_dimensions(note)
    anb_char = len(self.note_text(note))
    tx = x1 + 5 + anb_char * self.canvas.char_h_size
    ty1 = (y1 + ty2) // 2 - 1
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(tx, ty1)
    while tx < x2 - 20:
      ctx.curve_to(
        tx + 10, ty1 - 5,
        tx + 10, ty1 + 5,
        tx + 20, ty1,
        )
      tx += 20
    ctx.stroke()
    
  def draw_note_fx_dead(self, ctx, note, string_id):
    x1, cy1, x2, cy2 = self.note_dimensions(note)
    anb_char = len(self.note_text(note))
    cx1 = x1 + 3
    cx2 = cx1 + anb_char * self.canvas.char_h_size
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(0.6)
    ctx.move_to(cx1, cy1)
    ctx.line_to(cx2, cy2)
    ctx.move_to(cx2, cy1)
    ctx.line_to(cx1, cy2)
    ctx.stroke()
    ctx.set_line_width(1)
    
  def draw_note_fx_roll(self, ctx, note, string_id):
    notes = note.chord_notes()
    min_string_id = 99999
    if notes:
      for other in notes:
        other_string_id = self.note_string_id(other)
        if other_string_id >     string_id: return
        if other_string_id < min_string_id: min_string_id = other_string_id
    else:
      min_string_id = self.note_string_id(note) - 1
    x1, y1, x2, y2 = self.note_dimensions(note)
    anb_char = len(self.note_text(note))
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1.0)
    ctx.move_to(x1 + 3 + anb_char * self.canvas.char_h_size, (y1 + y2) // 2)
    ctx.rel_curve_to(
      10,  0,
      16, -(string_id - min_string_id) * self.string_height + 20,
      16, -(string_id - min_string_id) * self.string_height,
      )
    ctx.rel_line_to(-5,  5)
    ctx.rel_move_to( 5, -5)
    ctx.rel_line_to( 5,  5)
    ctx.stroke()
    
    
  def _nothing(self, ctx, x, y1, y2, note, previous, next): pass
  _round = _nothing
  
  def _white(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.8, 0.8, 0.8)
    ctx.set_line_width(2)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    ctx.set_line_width(1)
  
  def _black(self, ctx, x, y1, y2, note, previous, next):
    #d = cmp(y2, y1) * 3
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    #ctx.line_to(x, y1 + d)
    ctx.line_to(x, y1)
    ctx.stroke()
  
#   def _quarter(self, ctx, x, y1, y2, time, duration, previous_time, next_time):
#     print time, previous_time, next_time
#     dx = time % 96
#     if not dx: dx  = -10
#     else:      dx -=  10
#     ctx.set_source_rgb(0.0, 0.0, 0.0)
#     ctx.set_line_width(1)
#     ctx.move_to(x, y2)
#     ctx.line_to(x, y1)
#     ctx.stroke()
#     ctx.set_line_width(2)
#     ctx.move_to(x, y1)
#     ctx.line_to(x - dx * self.canvas.zoom, y1)
#     ctx.stroke()
#     ctx.set_line_width(1)

#   def _hquarter(self, ctx, x, y1, y2, time, duration, previous_time, next_time):
#     d = cmp(y2, y1) * 4
#     dx2 = time % 48
#     if not dx2: dx2 = -8
#     else:       dx2 -= 8
#     dx = time % 96
#     if not dx: dx = -10
#     else:      dx = min(24, dx - 10)
#     ctx.set_source_rgb(0.0, 0.0, 0.0)
#     ctx.set_line_width(1)
#     ctx.move_to(x, y2)
#     ctx.line_to(x, y1)
#     ctx.stroke()
#     ctx.set_line_width(2)
#     ctx.move_to(x, y1)
#     ctx.line_to(x - dx  * self.canvas.zoom, y1)
#     ctx.move_to(x, y1 + d)
#     ctx.line_to(x - dx2 * self.canvas.zoom, y1 + d)
#     ctx.stroke()
#     ctx.set_line_width(1)
    
#   def _hhquarter(self, ctx, x, y1, y2, time, duration, previous_time, next_time):
#     d = cmp(y2, y1) * 4
#     dx3 = time % 24
#     if not dx3: dx3 = -6
#     else:       dx3 -= 6
#     dx2 = time % 48
#     if not dx2: dx2 = -8
#     else:       dx2 = min(12, dx2 - 8)
#     dx = time % 96
#     if not dx: dx = -10
#     else:      dx = min(12, dx - 10)
#     ctx.set_source_rgb(0.0, 0.0, 0.0)
#     ctx.set_line_width(1)
#     ctx.move_to(x, y2)
#     ctx.line_to(x, y1)
#     ctx.stroke()
#     ctx.set_line_width(2)
#     ctx.move_to(x, y1)
#     ctx.line_to(x - dx  * self.canvas.zoom, y1)
#     ctx.move_to(x, y1 + d)
#     ctx.line_to(x - dx2 * self.canvas.zoom, y1 + d)
#     ctx.move_to(x, y1 + d * 2)
#     ctx.line_to(x - dx3 * self.canvas.zoom, y1 + d * 2)
#     ctx.stroke()
#     ctx.set_line_width(1)

  def _quarter(self, ctx, x, y1, y2, note, previous, next):
    #print time, previous_time, next_time
    
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    dt           = note.time % 96
    has_next     = next     and (next.time + next.duration <= note.time - dt + 96) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.time >= note.time - dt) and (previous.time + previous.duration >= note.time)
    if has_next:
      if note.time + note.duration < note.time - dt + 96:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      if previous.duration > 48 * 1.5:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x - (note.time - previous.time) * self.canvas.zoom, y1)
        ctx.stroke()
        ctx.set_line_width(1)
      return
    # Single
    sens = cmp(y2, y1)
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x    , y1 + 2 * sens)
    ctx.fill()
    
  def _hquarter(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    dt   = note.time % 96
    sens = cmp(y2, y1)
    has_next     = next     and (next.time + next.duration <= note.time - dt + 96) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.time >= note.time - dt) and (previous.time + previous.duration == note.time)
    if has_next:
      if note.time + note.duration < note.time - dt + 96:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        #if (not has_previous) or (note.time + note.duration < note.time - dt % 48 + 48):
        if   (next.duration <= 24 * 1.5):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 24 * 1.5)):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      if (previous.duration > 24 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 3 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      return
    
    # Single
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x, y1 + 2 * sens)
    ctx.curve_to(x + 3, y1 +  7 * sens, x + 7, y1 +  6 * sens, x + 5, y1 + 20 * sens)
    ctx.curve_to(x + 7, y1 + 11 * sens, x + 2, y1 +  9 * sens, x, y1 +  8 * sens)
    ctx.fill()
    
  def _hhquarter(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    dt   = note.time % 96
    sens = cmp(y2, y1)
    has_next     = next     and (next.time + next.duration <= note.time - dt + 96) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.time >= note.time - dt) and (previous.time + previous.duration == note.time)
    if has_next:
      if note.time + note.duration < note.time - dt + 96:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        #if (not has_previous) or (note.time + note.duration < note.time - dt % 48 + 48):
        if (next.duration <= 24 * 1.5):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 24 * 1.5)):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        #if (not has_previous) or (note.time + note.duration < note.time - dt % 24 + 24):
        if (next.duration <= 12 * 1.5):
          ctx.move_to(x, y1 + 6 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 6 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 12 * 1.5)):
          ctx.move_to(x, y1 + 6 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 6 * sens)
          ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      if (previous.duration > 24 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 3 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      if (previous.duration > 12 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 6 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 6 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      return

    # Single
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x, y1 + 2 * sens)
    ctx.curve_to(x + 3, y1 +  7 * sens, x + 7, y1 +  6 * sens, x + 5, y1 + 20 * sens)
    ctx.curve_to(x + 7, y1 + 11 * sens, x + 2, y1 +  9 * sens, x, y1 +  8 * sens)
    ctx.curve_to(x + 3, y1 + 13 * sens, x + 7, y1 + 12 * sens, x + 5, y1 + 26 * sens)
    ctx.curve_to(x + 7, y1 + 17 * sens, x + 2, y1 + 15 * sens, x, y1 + 14 * sens)
    ctx.fill()
    
  def _black_triplet(self, ctx, x, y1, y2, note, previous, next):
    sens = cmp(y2, y1)
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1 + 3 * sens)
    ctx.stroke()
    
    dt = note.time % 192
    
    if dt == 64:
      if    y2 > y1: ctx.move_to(x - self.canvas.char_h_size // 2, y1 - self.canvas.default_descent)
      else:          ctx.move_to(x - self.canvas.char_h_size // 2, y1 + self.canvas.default_ascent)
      ctx.set_font_size(self.canvas.default_font_size // 2)
      ctx.show_text("3")
      ctx.set_font_size(self.canvas.default_font_size)
      
      ctx.move_to(x - 64 * self.canvas.zoom, y1 + 0.5)
      ctx.line_to(x + 64 * self.canvas.zoom, y1 + 0.5)
      ctx.stroke()
      
  def _quarter_triplet(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    dt           = note.time % 96
    has_next     = next     and (next.time + next.duration <= note.time - dt + 96) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.time >= note.time - dt) and (previous.time + previous.duration == note.time)
    
    if dt == 32:
      if    y2 > y1: ctx.move_to(x - self.canvas.char_h_size // 2, y1 - self.canvas.default_descent)
      else:          ctx.move_to(x - self.canvas.char_h_size // 2, y1 + self.canvas.default_ascent)
      ctx.set_font_size(self.canvas.default_font_size // 2)
      ctx.show_text("3")
      ctx.set_font_size(self.canvas.default_font_size)
      
    if has_next:
      if note.time + note.duration < note.time - dt + 96:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      return
    # Single
    sens = cmp(y2, y1)
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x    , y1 + 2 * sens)
    ctx.fill()
      
  def _hquarter_triplet(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    if (note.time % 48) == 16:
      if    y2 > y1: ctx.move_to(x - self.canvas.char_h_size // 2, y1 - self.canvas.default_descent)
      else:          ctx.move_to(x - self.canvas.char_h_size // 2, y1 + self.canvas.default_ascent)
      ctx.set_font_size(self.canvas.default_font_size // 2)
      ctx.show_text("3")
      ctx.set_font_size(self.canvas.default_font_size)
      
    dt   = note.time % 96
    sens = cmp(y2, y1)
    has_next     = next     and (next.time + next.duration <= note.time - dt + 96) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.time >= note.time - dt) and (previous.time + previous.duration == note.time)
    if has_next:
      if note.time + note.duration < note.time - dt + 96:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        if   (next.duration <= 24 * 1.5):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 24 * 1.5)):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      if (previous.duration > 24 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 3 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      return
    
    # Single
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x, y1 + 2 * sens)
    ctx.curve_to(x + 3, y1 +  7 * sens, x + 7, y1 +  6 * sens, x + 5, y1 + 20 * sens)
    ctx.curve_to(x + 7, y1 + 11 * sens, x + 2, y1 +  9 * sens, x, y1 +  8 * sens)
    ctx.fill()
    
  def _hhquarter_triplet(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    if (note.time % 24) == 8:
      if    y2 > y1: ctx.move_to(x - self.canvas.char_h_size // 2, y1 - self.canvas.default_descent)
      else:          ctx.move_to(x - self.canvas.char_h_size // 2, y1 + self.canvas.default_ascent)
      ctx.set_font_size(self.canvas.default_font_size // 2)
      ctx.show_text("3")
      ctx.set_font_size(self.canvas.default_font_size)
      
    dt   = note.time % 96
    sens = cmp(y2, y1)
    has_next     = next     and (next.time + next.duration <= note.time - dt + 96) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.time >= note.time - dt) and (previous.time + previous.duration == note.time)
    if has_next:
      if note.time + note.duration < note.time - dt + 96:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        #if (not has_previous) or (note.time + note.duration < note.time - dt % 48 + 48):
        if (next.duration <= 24 * 1.5):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 24 * 1.5)):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        #if (not has_previous) or (note.time + note.duration < note.time - dt % 24 + 24):
        if (next.duration <= 12 * 1.5):
          ctx.move_to(x, y1 + 6 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 6 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 12 * 1.5)):
          ctx.move_to(x, y1 + 6 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 6 * sens)
          ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      if (previous.duration > 24 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 3 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      if (previous.duration > 12 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 6 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 6 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      return

    # Single
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x, y1 + 2 * sens)
    ctx.curve_to(x + 3, y1 +  7 * sens, x + 7, y1 +  6 * sens, x + 5, y1 + 20 * sens)
    ctx.curve_to(x + 7, y1 + 11 * sens, x + 2, y1 +  9 * sens, x, y1 +  8 * sens)
    ctx.curve_to(x + 3, y1 + 13 * sens, x + 7, y1 + 12 * sens, x + 5, y1 + 26 * sens)
    ctx.curve_to(x + 7, y1 + 17 * sens, x + 2, y1 + 15 * sens, x, y1 + 14 * sens)
    ctx.fill()
  
  def _quarter8(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    dt           = note.time % 144
    has_next     = next     and (next    .duration <= 48 * 1.5) and (next.time + next.duration <= note.time - dt + 144) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.duration <= 48 * 1.5) and (previous.time >= note.time - dt) and (previous.time + previous.duration == note.time)
    if has_next:
      if note.time + note.duration < note.time - dt + 144:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      return
    # Single
    sens = cmp(y2, y1)
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x    , y1 + 2 * sens)
    ctx.fill()
    
  def _hquarter8(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    sens = cmp(y2, y1)
    dt           = note.time % 144
    has_next     = next     and (next    .duration <= 48 * 1.5) and (next.time + next.duration <= note.time - dt + 144) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.duration <= 48 * 1.5) and (previous.time >= note.time - dt) and (previous.time + previous.duration == note.time)
    if has_next:
      if note.time + note.duration < note.time - dt + 144:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        if (next.duration <= 24 * 1.5):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 24 * 1.5)):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      if (previous.duration > 24 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 3 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      return
    
    # Single
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x, y1 + 2 * sens)
    ctx.curve_to(x + 3, y1 +  7 * sens, x + 7, y1 +  6 * sens, x + 5, y1 + 20 * sens)
    ctx.curve_to(x + 7, y1 + 11 * sens, x + 2, y1 +  9 * sens, x, y1 +  8 * sens)
    ctx.fill()
  
  def _hhquarter8(self, ctx, x, y1, y2, note, previous, next):
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(x, y2)
    ctx.line_to(x, y1)
    ctx.stroke()
    
    sens = cmp(y2, y1)
    dt           = note.time % 144
    has_next     = next     and (next    .duration <= 48 * 1.5) and (next.time + next.duration <= note.time - dt + 144) and (note.time + note.duration == next.time)
    has_previous = previous and (previous.duration <= 48 * 1.5) and (previous.time >= note.time - dt) and (previous.time + previous.duration == note.time)
    if has_next:
      if note.time + note.duration < note.time - dt + 144:
        ctx.set_line_width(2)
        ctx.move_to(x, y1)
        ctx.line_to(x + note.duration * self.canvas.zoom, y1)
        ctx.stroke()
        if (next.duration <= 24 * 1.5):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 24 * 1.5)):
          ctx.move_to(x, y1 + 3 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
          ctx.stroke()
        if (next.duration <= 12 * 1.5):
          ctx.move_to(x, y1 + 6 * sens)
          ctx.line_to(x + note.duration * self.canvas.zoom, y1 + 6 * sens)
          ctx.stroke()
        elif not (has_previous and (previous.duration <= 12 * 1.5)):
          ctx.move_to(x, y1 + 6 * sens)
          ctx.line_to(x + note.duration // 2 * self.canvas.zoom, y1 + 6 * sens)
          ctx.stroke()
        ctx.set_line_width(1)
      return
    if has_previous:
      if (previous.duration > 24 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 3 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 3 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      if (previous.duration > 12 * 1.5): # Else, previous will draw the stem
        ctx.set_line_width(2)
        ctx.move_to(x, y1 + 6 * sens)
        ctx.line_to(x - note.duration // 2 * self.canvas.zoom, y1 + 6 * sens)
        ctx.stroke()
        ctx.set_line_width(1)
      return
    
    # Single
    ctx.move_to(x, y1)
    ctx.curve_to(x + 3, y1 + sens, x + 7, y1, x + 5, y1 + 14 * sens)
    ctx.curve_to(x + 7, y1 + 5 * sens, x + 2, y1 + 3 * sens, x, y1 + 2 * sens)
    ctx.curve_to(x + 3, y1 +  7 * sens, x + 7, y1 +  6 * sens, x + 5, y1 + 20 * sens)
    ctx.curve_to(x + 7, y1 + 11 * sens, x + 2, y1 +  9 * sens, x, y1 +  8 * sens)
    ctx.curve_to(x + 3, y1 + 13 * sens, x + 7, y1 + 12 * sens, x + 5, y1 + 26 * sens)
    ctx.curve_to(x + 7, y1 + 17 * sens, x + 2, y1 + 15 * sens, x, y1 + 14 * sens)
    ctx.fill()

    
  _drawers = { # Maps durations to the corresponding drawer function
    576 : _round,
    288 : _white,
    144 : _black,
    72  : _quarter,
    36  : _hquarter,
    18  : _hhquarter,

    384 : _round,
    192 : _white,
    96  : _black,
    48  : _quarter,
    24  : _hquarter,
    12  : _hhquarter,

    64  : _black_triplet,
    32  : _quarter_triplet,
    16  : _hquarter_triplet,
    8   : _hhquarter_triplet,
    }
  _drawers8 = { # Maps durations to the corresponding drawer function for x/8 rythms
    576 : _round,
    288 : _white,
    144 : _black,
    72  : _quarter8,
    36  : _hquarter8,
    18  : _hhquarter8,
    
    384 : _round,
    192 : _white,
    96  : _black,
    48  : _quarter8,
    24  : _hquarter8,
    12  : _hhquarter8,
    
    # Does triplets make sense in 6/8 rythms ???
    64  : _black_triplet,
    32  : _quarter_triplet,
    16  : _hquarter_triplet,
    8   : _hhquarter_triplet,
    }
  
  def draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure = None, previous = None, next = None):
    if mesure and (mesure.rythm2 == 8): return (PartitionDrawer._drawers8.get(note.duration) or PartitionDrawer._nothing)(self, ctx, x, y1, y2, note, previous, next)
    else:                               return (PartitionDrawer._drawers .get(note.duration) or PartitionDrawer._nothing)(self, ctx, x, y1, y2, note, previous, next)
    
    
    
    
class TablatureDrawer(PartitionDrawer):
  def __init__(self, canvas, partition):
    PartitionDrawer.__init__(self, canvas, partition)
    self.height = 300
    self.strings = partition.view.strings
    self.string_height = canvas.default_line_height + 2
    self.stem_extra_height = self.string_height
    self.stem_extra_height_bottom = self.string_height * 0.6
    
  def note_string_id(self, note):
    if not hasattr(note, "string_id"):
      note.string_id = self.default_string_id(note.value)
    return min(note.string_id, len(self.strings) - 1)
  
  def note_text(self, note):
    if note is self.canvas.cursor: return "_"
    return self.strings[self.note_string_id(note)].value_2_text(note.value)
  
  def default_string_id(self, value):
    value = abs(value)
    for i, string in enumerate(self.strings):
      if string.base_note <= value: return i
    return len(self.strings) - 1
    
  def draw_note_fx_link(self, ctx, note, string_id):
    ax1, ay1, ax2, ay2 = self.note_dimensions(note)
    if note.linked_to: bx1, by1, bx2, by2 = self.note_dimensions(note.linked_to)
    else:              bx1, by1, bx2, by2 = ax1 + note.duration, ay1, ax2 + note.duration, ay2
    anb_char = len(self.note_text(note))
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    ctx.set_line_width(1)
    ctx.move_to(ax1 + 5 + self.canvas.char_h_size * anb_char, (ay1 + ay2) // 2 + self.canvas.default_descent)
    ctx.curve_to(
      ax1 + 8 + self.canvas.char_h_size * anb_char, (ay1 + ay2) // 2 + self.canvas.default_descent + 4,
      bx1 - 1                                     , (by1 + by2) // 2 + self.canvas.default_descent + 4,
      bx1 + 2                                     , (by1 + by2) // 2 + self.canvas.default_descent,
      )
    ctx.stroke()
    if note.linked_to:
      if   note.value < note.linked_to.value:
        ctx.move_to((ax1 + self.canvas.char_h_size * anb_char + bx1) // 2, (ay1 + ay2 + by1 + by2) // 4 + self.canvas.default_descent)
        ctx.show_text("h")
      elif note.value > note.linked_to.value:
        ctx.move_to((ax1 + self.canvas.char_h_size * anb_char + bx1) // 2, (ay1 + ay2 + by1 + by2) // 4 + self.canvas.default_descent)
        ctx.show_text("p")
    
  def draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure = None, previous_time = -32000, next_time = 32000):
    x = x + self.string_height // 2.5
    PartitionDrawer.draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure, previous_time, next_time)
    
    
class DrumsDrawer(PartitionDrawer):
  def __init__(self, canvas, partition):
    PartitionDrawer.__init__(self, canvas, partition)
    self.height = 300
    self.strings = partition.view.strings
    self.string_height = canvas.default_line_height + 2
    self.stem_extra_height = self.string_height
    
  def note_string_id(self, note):
    for i, string in enumerate(self.strings):
      if string.base_note == abs(note.value): return i
    return -1
  
  def note_text(self, note):
    if note is self.canvas.cursor: return "_"
    return "O"
  
  def default_string_id(self, value):
    value = abs(value)
    for i, string in enumerate(self.strings):
      if string.base_note == value: return i
    return -1
    
  def draw(self, ctx, x, y, width, height, drag = 0):
    PartitionDrawer.draw(self, ctx, x, y, width, height, drag)
    if not drag:
      ctx.set_source_rgb(0.4, 0.4, 0.4)
      ctx.set_font_size(self.canvas.default_font_size * 0.8)
      for i in range(len(self.strings)):
        ctx.move_to(self.canvas.start_x + 3, self.string_id_2_y(i))
        ctx.show_text(model.DRUM_PATCHES[self.strings[i].base_note])
        
      ctx.set_source_rgb(0.0, 0.0, 0.0)
      ctx.set_font_size(self.canvas.default_font_size)
      
  def draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure = None, previous_time = -32000, next_time = 32000):
    x = x + self.string_height // 2.5
    PartitionDrawer.draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure, previous_time, next_time)

STAFF_STRING_PREVIOUS = { 0 : -1, 2 : 0, 4 : 2, 5 : 4, 7 : 5, 9 :  7, 11 :  9 }
STAFF_STRING_NEXT     = { 0 :  2, 2 : 4, 4 : 5, 5 : 7, 7 : 9, 9 : 11, 11 : 12 }

class StaffDrawer(PartitionDrawer):
  def __init__(self, canvas, partition, compact = 0):
    PartitionDrawer.__init__(self, canvas, partition)
    self.compact = compact
    if compact:
      self.string_height  = canvas.default_line_height // 7 + 2
    else:
      self.string_height  = canvas.default_line_height // 4 + 2
    self.stem_extra_height        = 15 + self.string_height
    self.stem_extra_height_bottom =  3 + self.string_height
    self.up_stem_ratio  = 1.0
    self.height         = 300
    self.strings        = []
    self.value_2_string = {}
    if getattr(partition, "g8", 0):
      string_values_and_line_styles = ((67, 0), (65, 2), (64, 0), (62, 2), (60, 0), (59, 2), (57, 0), (55, 2), (53, 0), (52, 2), (50, 0))
    else:
      string_values_and_line_styles = ((79, 0), (77, 2), (76, 0), (74, 2), (72, 0), (71, 2), (69, 0), (67, 2), (65, 0), (64, 2), (62, 0))
      
    alterations = model.TONALITIES[partition.tonality]
    if alterations:
      if alterations[0] == model.DIESES[0]: self.alteration_type =  1 # diese
      else:                                 self.alteration_type = -1 # bemol
    else:                                   self.alteration_type =  0
    self.alterations  = set(alterations)
    alterations2      = set(alterations)
    
    for note_value, line_style in string_values_and_line_styles:
      alteration      = 0
      draw_alteration = 0
      if (note_value % 12) in self.alterations:
        alteration = self.alteration_type
        if (note_value % 12) in alterations2:
          alterations2.remove(note_value % 12)
          draw_alteration = 1
      StaffString(self, note_value, line_style, alteration, draw_alteration)
    self._sort_strings()
    
    self.key_string = self.strings[7] # G string
    
  def draw_strings(self, ctx, x, y, width, height):
    if self.compact:
      draw_bitmap(self.canvas, ctx,
                  os.path.join(globdef.DATADIR, "clef_de_sol_ps.png"),
                  2,
                  int(self.string_id_2_y(self.strings.index(self.key_string)) - .5 * 112 * 0.62),
                  0.5)
    else:
      draw_bitmap(self.canvas, ctx,
                  os.path.join(globdef.DATADIR, "clef_de_sol.png"),
                  2,
                  int(self.string_id_2_y(self.strings.index(self.key_string)) - 85 * .62),
                  1.0)
    PartitionDrawer.draw_strings(self, ctx, 0, y, width + x, height)
    #if self.height < 130: self.height = 130
    
    
  def draw(self, ctx, x, y, width, height, drag = 0):
    PartitionDrawer.draw(self, ctx, x, y, width, height, drag, 0)
    if not drag:
      y = self.y + self.start_y + self.stem_extra_height + self.canvas.default_line_height // 4 #+ self.string_height // 2
      
      for string in self.strings:
        y += self.string_height
        if string.draw_alteration:
          if string.alteration == -1:
            ctx.move_to(30 + self.canvas.char_h_size * 0.35 * model.BEMOLS.index(string.base_note % 12), y - self.canvas.default_line_height // 5)
            ctx.show_text(u"♭")
          else:
            ctx.move_to(30 + self.canvas.char_h_size * 0.35 * model.DIESES.index(string.base_note % 12), y)
            ctx.show_text(u"♯")
      
  def _sort_strings(self):
    self.strings.sort(lambda a, b: cmp(b.base_note, a.base_note))
    for i, string in enumerate(self.strings): string.id = i
    
  def get_string(self, note_value, create = 1):
    if note_value > self.strings[0].base_note:
      if create:
        new_string_value = STAFF_STRING_NEXT[self.strings[0].base_note % 12] + (self.strings[0].base_note // 12) * 12
        if self.strings[0].line_style == 0: new_string_line_style = 1
        else:                               new_string_line_style = 0
        if (new_string_value % 12) in self.alterations: alteration = self.alteration_type
        else:                                           alteration = 0
        string = StaffString(self, new_string_value, new_string_line_style, alteration, 0)
        self._sort_strings()
        self.canvas.render_all()
        return self.get_string(note_value, create)
      else:
        return None
      
    string = self.value_2_string.get(note_value)
    if string: return string
    #for string in self.strings:
    #  if note_value >= string.base_note: return string
      
    if create:
      new_string_value = STAFF_STRING_PREVIOUS[self.strings[-1].base_note % 12] + (self.strings[-1].base_note // 12) * 12
      if self.strings[-1].line_style == 0: new_string_line_style = 1
      else:                                new_string_line_style = 0
      if (new_string_value % 12) in self.alterations: alteration = self.alteration_type
      else:                                           alteration = 0
      string = StaffString(self, new_string_value, new_string_line_style, alteration, 0)
      self._sort_strings()
      self.canvas.render_all()
      return self.get_string(note_value, create)

  def note_string_id(self, note):
    unaltered_value, alteration = note.unaltered_value_and_alteration()
    return self.get_string(unaltered_value, 1).id
  
  def note_text(self, note):
    if note is self.canvas.cursor: return "-"
    return self.get_string(note.unaltered_value_and_alteration()[0]).value_2_text(note.value)
  
  def note_dimensions(self, note):
    string_id = self.note_string_id(note)
    x = self.canvas.time_2_x(note.time)
    #if abs(note.value) == self.strings[string_id].base_note: x1 = x
    #else:                                                    x1 = x - 5 - self.canvas.char_h_size 
    y = self.string_id_2_y(string_id) - self.string_height
    return (
      x,
      y,
      x + note.duration * self.canvas.zoom,
      y + self.canvas.default_line_height // 2 + 4,
      )
  
  def note_previous(self, note):
    string_id = self.note_string_id(note)
    while 1:
      note = note.previous()
      if not note: return None
      if self.note_string_id(note) == string_id: return note
      
  def render_note(self, note):
    mesure = self.canvas.song.mesure_at(note)
    if mesure.rythm2 == 8: time1 = (note.time // 144) * 144; time2 = max(note.end_time(), time1 + 144)
    else:                  time1 = (note.time //  96) *  96; time2 = max(note.end_time(), time1 +  96)
    
    # Required for updating sharp / natural / flat symbol correctly
    if self.partition.notes:
      time2 = max(time2, self.partition.notes[-1].end_time())
      
    extra_width = 3 + self.canvas.char_h_size * 0.8 # for # and b
    self.canvas.render_pixel(
      self.canvas.time_2_x(time1) - extra_width,
      self.y + self.start_y - 1,
      (time2 - time1) * self.canvas.zoom + 2 + extra_width,
      self.height - self.start_y + 3,
      )
    
  def draw_note(self, ctx, note, string_id, y):
    x = self.canvas.time_2_x(note.time)
    ctx.translate(x + 6.5, y + 0.5)
    ctx.rotate(-0.3)
    ctx.scale(1.0, 0.7)
    if note.value < 0: ctx.set_source_rgb(0.7, 0.7, 0.7)
    else: self.note_color(ctx, note)

    if note.fx != "dead":
      white = note.base_duration() in (192, 384)
      if white:
        ctx.arc(0, 0, 4.0, 0, 2.0 * math.pi)
        ctx.set_line_width(2.0)
        ctx.stroke()
        ctx.set_line_width(1.0)
      else:
        ctx.arc(0, 0, 5.0, 0, 2.0 * math.pi)
        ctx.fill()
    ctx.identity_matrix()

    alteration = abs(note.value) - self.strings[string_id].base_note
    
    previous = self.note_previous(note)
    if previous:
      previous_alteration = abs(previous.value) - self.strings[string_id].base_note
      if self.partition.song.mesure_at(previous) is not self.partition.song.mesure_at(note):
        if previous_alteration != self.strings[string_id].alteration:
          previous_alteration = -99 # force the drawing of the alteration mark
    else:
      previous_alteration = self.strings[string_id].alteration
      
    if alteration != previous_alteration:
      ctx.move_to(x - self.canvas.char_h_size, y + self.string_height + 3)
      if   alteration == -2: ctx.show_text(u"♭♭")
      elif alteration == -1: ctx.show_text(u"♭")
      elif alteration ==  0: ctx.show_text(u"♮")
      elif alteration ==  1: ctx.show_text(u"♯")
      elif alteration ==  2: ctx.show_text(u"♯♯")
      ctx.new_path()
      
    if note.is_dotted():
      ctx.arc(x + 13, y + self.string_height / 2 + 2, 1.5, 0, 2.0 * math.pi)
      ctx.fill()
      
    if note.fx: getattr(self, "draw_note_fx_%s" % note.fx)(ctx, note, string_id)
    ctx.set_source_rgb(0.0, 0.0, 0.0)
    
  def draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure = None, previous_time = -32000, next_time = 32000):
    if y1 < y2:
      x  += 11
      y2 += 2
    else:
      x  += 1
      y2 -= 2
    PartitionDrawer.draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure, previous_time, next_time)
    

class StaffString(object):
  def __init__(self, drawer, base_note, line_style, alteration, draw_alteration):
    self.drawer              = drawer
    self.base_note           = base_note
    self.line_style          = line_style
    self.alteration          = alteration
    self.draw_alteration     = draw_alteration
    
    drawer.value_2_string[base_note] = self
    drawer.strings.append(self)
    
  def value_2_text(self, value): return u"o"
  def text_2_value(self, text):  return self.base_note + self.alteration
  
  def __unicode__(self): return _(self.__class__.__name__) % note_label(self.base_note)
  
  def width(self):
    if   self.line_style == 0: return -1
    elif self.line_style == 1: return 8
    return 9
  

class LyricsDrawer(Drawer):
  def __init__(self, canvas, lyrics):
    Drawer.__init__(self, canvas)
    
    self.last_undoable = None
    self.lyrics  = lyrics
    self.x0      = canvas.start_x + 2
    self.start_y = 0
    self.melody  = None
    self.timeout = None
    self.text    = gtk.TextView()
    self.text.get_buffer().set_text(lyrics.text)
    self.update_melody()
    self.canvas.partition_2_drawer[lyrics] = self
    
    self.fixed = gtk.Fixed()
    self.fixed.set_has_window(1)
    self.fixed.connect("button_press_event", self.on_button_press)
    self.fixed.put(self.text, 0, 1)
    canvas.put(self.fixed, canvas.start_x + 2, 0)
    
    self.text.connect("focus-in-event", self.on_text_focus)
    self.text.connect("key_press_event", self.on_text_key_press)
    self.text_changed_handler = self.text.get_buffer().connect("changed", self.on_text_changed)
    self.text.show_all()
    canvas   .show_all()
    
  def partition_listener(self, partition, type, new, old): pass
    
  def on_text_changed(self, buffer):
    old_text = self.lyrics.text
    new_text = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter())
    
    def set_text(text):
      current = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter())
      self.lyrics.text = text
      if current != text:
        buffer.handler_block(self.text_changed_handler)
        buffer.set_text(text)
        buffer.handler_unblock(self.text_changed_handler)
        self.set_text_size() # Not delayed, because the set_text remove all the precendent text tags !
      else:
        self.delayed_set_text_size()
        
    def do_it  (): set_text(new_text)
    def undo_it(): set_text(old_text)
    
    if self.last_undoable and self.canvas.main.undo_stack.undoables and (self.last_undoable is self.canvas.main.undo_stack.undoables[-1]):
      self.last_undoable.do_func = do_it
      do_it()
    else:
      self.last_undoable = UndoableOperation(do_it, undo_it, _(u"edit lyrics"), self.canvas.main.undo_stack)
      
  def on_text_focus(self, text, event):
    if getattr(self.canvas.main, "notebook", None):
      self.canvas.main.notebook.edit(-3, self.lyrics)
      
  def on_text_key_press(self, text, event):
    if    (event.keyval == 32) and (event.state == 0):
      text.get_buffer().insert_at_cursor("\t")
      return 1
    elif ((event.keyval == 45) or (event.keyval == 65455)) and (event.state == 0):
      text.get_buffer().insert_at_cursor("-\t")
      return 1
    
  def melody_note_listener(self, obj, type, new, old):
    self.update_melody()
    
  def update_melody(self):
    melody = self.lyrics.get_melody()
    if melody is not self.melody:
      if self.melody: unobserve(self.melody.notes, self.melody_note_listener)
      self.melody = melody
      if self.melody:   observe(self.melody.notes, self.melody_note_listener)

    self.sizes = []
    if melody and melody.notes:
      tabs = pango.TabArray(len(melody.notes) - 1, 1)
      previous_time = melody.notes[0].time
      i = 0
      x = 0
      for note in melody.notes[1:]:
        dx = ((note.time - previous_time) * self.canvas.zoom)
        x += dx
        self.sizes.append(dx)
        tabs.set_tab(i, pango.TAB_LEFT, int(x))
        i += 1
        previous_time = note.time
      self.text.set_tabs(tabs)
      
      #x0 = int(self.canvas.time_2_x(melody.notes[0].time)) + 2 - self.canvas.start_x
      x0 = int(2 + melody.notes[0].time * self.canvas.zoom)
      if self.x0 != x0:
        self.x0 = x0
        self.canvas.render_all()
        
    self.delayed_set_text_size()

  def delayed_set_text_size(self):
    if self.timeout: gobject.source_remove(self.timeout)
    self.timeout = gobject.timeout_add(120, self.set_text_size)
    
  def set_text_size(self):
    b = self.text.get_buffer()
    b.remove_all_tags(b.get_start_iter(), b.get_end_iter())
    
    layout = pango.Layout(self.text.get_pango_context())
    
    line_id = 0
    for line in self.lyrics.text.split(u"\n"):
      i      = 0
      offset = 0
      for piece in line.split(u"\t")[:-1]:
        if i >= len(self.sizes): continue
        
        layout.set_text(piece)
        width = layout.get_pixel_extents()[1][2]
        if width == 0: width = 1
        
        max_width = self.sizes[i] - 7.0
        if width > max_width:
          tag = b.create_tag(scale = max(0.01, max_width / width))
          b.apply_tag(tag, b.get_iter_at_line_offset(line_id, offset), b.get_iter_at_line_offset(line_id, offset + 1 + len(piece)))
          
        offset += 1 + len(piece)
        i      += 1
        
      line_id += 1
      
  def draw(self, ctx, x, y, width, height, drag = 0):
    if Drawer.draw(self, ctx, x, y, width, height, drag):
      self.start_y = draw_multiline_text(ctx, self.lyrics.header, 0, self.y + 2, self.canvas.width - 20) - self.y + 7
      
      text_w, text_h = self.text.size_request()
      self.height = self.start_y + max(editor_gtk.SMALL_ICON_SIZE, text_h + 2)
      if self.canvas.x == 0:
        ctx.set_source_rgb(0.0, 0.0, 0.0)
        ctx.set_line_width(2)
        ctx.move_to(self.canvas.start_x - self.canvas.x + 1, self.y + self.start_y)
        ctx.rel_line_to(0, self.height - self.start_y)
        ctx.stroke()
        ctx.set_line_width(1)
        
      # Draws the icon
      pixbuf = editor_gtk.load_small_icon(description(self.lyrics.__class__).icon_filename_for(self.lyrics))
      self.canvas.canvas.window.draw_pixbuf(None, pixbuf, 0, 0, 4, int(self.y + self.start_y // 2 + self.height // 2 - editor_gtk.SMALL_ICON_SIZE // 2), -1, -1)
      
    if self.canvas.child_get_property(self.fixed, "y") != self.y + self.start_y:
      self.canvas.child_set_property(self.fixed, "y", self.y + self.start_y)
      
    if self.fixed.child_get_property(self.text, "x") != self.x0 - self.canvas.x:
      self.fixed.child_set_property(self.text, "x", self.x0 - self.canvas.x)
      
    self.fixed.set_size_request(self.canvas.width, int(self.height - self.start_y))
    
  def on_button_press(self, *args): self.text.grab_focus()
  
  def destroy(self):
    self.fixed.destroy()
    Drawer.destroy(self)
    
def draw_round_rect(ctx, x1, y1, x2, y2, round):
  round = min(round, (x2 - x1) // 2, (y2 - y1) // 2)
  round2 = round / 5
  ctx.move_to(x2 - round, y1        ); ctx.curve_to(x2 - round2, y1         , x2         , y1 + round2, x2        , y1 + round)
  ctx.line_to(x2        , y2 - round); ctx.curve_to(x2         , y2 - round2, x2 - round2, y2         , x2 - round, y2        )
  ctx.line_to(x1 + round, y2        ); ctx.curve_to(x1 + round2, y2         , x1         , y2 - round2, x1        , y2 - round)
  ctx.line_to(x1        , y1 + round); ctx.curve_to(x1         , y1 + round2, x1 + round2, y1         , x1 + round, y1        )
  ctx.close_path()
  
  

def draw_multiline_text(ctx, text, x, y, max_width):
  ascent, descent, font_height, max_x_advance, max_y_advance = ctx.font_extents()
  lines = text.split(u"\n")
  for line in lines:
    x_bearing, y_bearing, width, height, x_advance, y_advance = ctx.text_extents(line)
    if width > max_width:
      words = line.split()
      words_ok = []
      line = u""
      for word in words:
        if line:
          line2 = line + u" " + word
          x_bearing, y_bearing, width, height, x_advance, y_advance = ctx.text_extents(line2)
          if width > max_width:
            y += font_height
            ctx.move_to(x, y)
            ctx.show_text(line)
            line = word
          else:
            line = line2
        else:
          line = word
          
      if line:
        y += font_height
        ctx.move_to(x, y)
        ctx.show_text(line)
        
    else:
      y += font_height
      ctx.move_to(x, y)
      ctx.show_text(line)
  return y

def draw_bitmap(canvas, ctx, bitmap_filename, x, y, scale = 1.0):
  if canvas.canvas:
    if scale < 1.0: pixbuf = editor_gtk.load_small_icon(bitmap_filename)
    else:           pixbuf = editor_gtk.load_big_icon  (bitmap_filename)
    canvas.canvas.window.draw_pixbuf(None, pixbuf, 0, 0, x, y, -1, -1)
  else:
    image = cairo.ImageSurface.create_from_png(bitmap_filename)
    #ctx.move_to(x, y)
    #ctx.save()
    ctx.scale(scale, scale)
    ctx.set_source_surface(image, x / scale, y / scale)
    ctx.paint()
    #ctx.restore()
    ctx.identity_matrix()

