# -*- coding: utf-8 -*-

# Songwrite 2
# Copyright (C) 2001-2007 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from songwrite2.main import *
from songwrite2.canvas import Canvas

import gobject, gtk

mainloop = gtk.main

sw_xml_filter = gtk.FileFilter()
sw_xml_filter.set_name(_(u"__songwrite_file_format__"))
sw_xml_filter.add_pattern(u"*.sw.xml")

MENU_LABEL_2_IMAGE = {
  u"Open..."        : gtk.STOCK_OPEN,
  u"Save"           : gtk.STOCK_SAVE,
  u"Save as..."     : gtk.STOCK_SAVE_AS,
  u"Close"          : gtk.STOCK_CLOSE,
  u"Undo"           : gtk.STOCK_UNDO,
  u"Redo"           : gtk.STOCK_REDO,
  u"Paste"          : gtk.STOCK_PASTE,
  u"Preferences..." : gtk.STOCK_PREFERENCES,
  }

icon_factory = gtk.IconFactory()
icon_factory.add_default()

class App(BaseApp, gtk.Window):
  def menu_image(self, filename):
    image = editobj2.editor_gtk.load_big_icon(os.path.join(globdef.DATADIR, filename))
    return image
  
  def append_to_menu(self, menu, has_submenu, label, command = None, arg = None, accel = "", image = None):
    if isinstance(menu, gtk.MenuItem):
      m = menu.get_submenu()
      if not m:
        m = gtk.Menu()
        menu.set_submenu(m)
      menu = m
    stock = MENU_LABEL_2_IMAGE.get(label)
    if stock:
      menu_item = gtk.ImageMenuItem(stock)
    elif image:
      gtk.stock_add([(label, _(label).replace("_", "__"), gtk.gdk.SHIFT_MASK, 46, "")])
      icon_factory.add(label, gtk.IconSet(image))
      menu_item = gtk.ImageMenuItem(label)
    else:
      menu_item = gtk.MenuItem(_(label).replace("_", "__"))
    if command:
      if arg: menu_item.connect("activate", command, arg)
      else:   menu_item.connect("activate", command)
    menu.append(menu_item)
    
    if accel:
      if accel.startswith(u"C-"):
        mod = gtk.gdk.CONTROL_MASK
      else:
        mod = 0
      key = accel[-1]
      menu_item.add_accelerator("activate", self.accel_group, ord(key), mod, gtk.ACCEL_VISIBLE)
      
    return menu_item
  
  def append_separator_to_menu(self, menu):
    menu.get_submenu().append(gtk.SeparatorMenuItem())

  def create_menubar(self):
    return gtk.MenuBar()
  
  def __init__(self, song = None):
    gtk.Window.__init__(self)
    
    self.accel_group = gtk.AccelGroup()
    self.add_accel_group(self.accel_group)
    self.set_role("main")
    
    self.vbox = gtk.VBox()
    self.add(self.vbox)

    BaseApp.__init__(self, song)

    toolbar = self.toolbar = gtk.Toolbar()
    toolbar.set_style(gtk.TOOLBAR_ICONS)
    #toolbar.set_property("icon-size", gtk.ICON_SIZE_MENU)
    toolbar.set_show_arrow(0)
    b = gtk.ToolButton(gtk.STOCK_SAVE)
    b.connect("clicked", self.on_save)
    toolbar.insert(b, -1)
    b = gtk.ToolButton(gtk.STOCK_PRINT)
    b.connect("clicked", self.on_preview_print)
    toolbar.insert(b, -1)
    #toolbar.insert(gtk.SeparatorToolItem(), -1)
    #b = gtk.ToolButton(gtk.STOCK_UNDO)
    #b.connect("clicked", self.on_undo)
    #toolbar.insert(b, -1)
    #b = gtk.ToolButton(gtk.STOCK_REDO)
    #b.connect("clicked", self.on_redo)
    #toolbar.insert(b, -1)
    toolbar.insert(gtk.SeparatorToolItem(), -1)
    b = gtk.ToolButton(gtk.STOCK_MEDIA_PLAY)
    b.connect("clicked", self.on_play)
    toolbar.insert(b, -1)
    b = gtk.ToolButton(gtk.STOCK_MEDIA_PLAY)
    b.connect("clicked", self.on_play_from_here)
    toolbar.insert(b, -1)
    b = gtk.ToolButton(gtk.STOCK_MEDIA_STOP)
    b.connect("clicked", self.on_stop_playing)
    toolbar.insert(b, -1)
    toolbar.insert(gtk.SeparatorToolItem(), -1)
    
    zoom_menubar = gtk.MenuBar()
    
    def change_zoom(menu_item, incr):
      gobject.timeout_add(100, zoom_menubar.cancel)
      self.canvas.change_zoom(incr)
      #i = zoom_levels.index(self.canvas.zoom) + incr
      #if   i < 0:                    i = 0
      #elif i > len(zoom_levels) - 1: i = len(zoom_levels) - 1
      #self.canvas.set_zoom(zoom_levels[i])
      #zoom_menu.get_child().set_text(u"%s%%" % int(self.canvas.zoom * 100.0))
      
    def set_zoom(menu_item, zoom):
      self.canvas.set_zoom(zoom)
      #zoom_menu.get_child().set_text(u"%s%%" % int(zoom * 100.0))
      
    zoom_out_menu = self.append_to_menu(zoom_menubar, 0, u" ‒ ")
    zoom_menu  = self.zoom_menu = self.append_to_menu(zoom_menubar, 1, u"100%")
    for zoom in songwrite2.canvas.zoom_levels: self.append_to_menu(zoom_menu, 0, u"%s%%" % int(zoom * 100.0), set_zoom, arg = zoom)
    zoom_in_menu  = self.append_to_menu(zoom_menubar, 0, u" + ")
    
    zoom_out_menu.connect("select", change_zoom, -1)
    zoom_in_menu .connect("select", change_zoom,  1)
    
    menu_box = gtk.HBox()
    self.vbox.pack_start(menu_box, 0, 1)

    menu_box.pack_start(self.menubar, 1, 1)
    menu_box.pack_start(toolbar     , 0, 1)
    menu_box.pack_end  (zoom_menubar, 0, 1)
    
    
    self.connect("delete-event", self.on_close)
    self.show_all()
    self.set_focus(self.canvas.canvas)
    observe(globdef.config, self.config_listener)
    
    
  def config_listener(self, obj, type, new, old):
    if type is object:
      if new["SIDE_BAR_POSITION"] != old["SIDE_BAR_POSITION"]: self.create_content()
    
  def create_content(self):
    if   self.floating: self.floating.destroy()
    if   self.paned:    self.paned   .destroy()
    elif self.scrolled: self.scrolled.destroy()
    
    if   (globdef.config.SIDE_BAR_POSITION == "left") or (globdef.config.SIDE_BAR_POSITION == "right") or (globdef.config.SIDE_BAR_POSITION == "floating_vertical"):
      self.notebook = EditNotebook(self.undo_stack, gtk.POS_TOP)
      self.notebook.set_current_page(self.notebook.get_n_pages() - 1)
    elif (globdef.config.SIDE_BAR_POSITION == "top") or (globdef.config.SIDE_BAR_POSITION == "bottom") or (globdef.config.SIDE_BAR_POSITION == "floating_horizontal"):
      self.notebook = EditNotebook(self.undo_stack, gtk.POS_LEFT)
      self.notebook.set_size_request(-1, 310)
    else:
      self.notebook = None
      
    if   (globdef.config.SIDE_BAR_POSITION == "left") or (globdef.config.SIDE_BAR_POSITION == "right"):
      self.paned    = gtk.HPaned()
      self.floating = None
    elif (globdef.config.SIDE_BAR_POSITION == "top") or (globdef.config.SIDE_BAR_POSITION == "bottom"):
      self.paned    = gtk.VPaned()
      self.floating = None
    elif (globdef.config.SIDE_BAR_POSITION == "floating_horizontal") or (globdef.config.SIDE_BAR_POSITION == "floating_vertical"):
      self.paned    = None
      self.floating = gtk.Window()
      self.floating.set_role("toolbox")
      self.floating.set_transient_for(self)
      self.floating.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_UTILITY)
      self.floating.set_destroy_with_parent(1)
    else:
      self.paned    = None
      self.floating = None
      
    if self.notebook:
      if self.songbook:
        self.songbook_pane = SongbookPane(self, self.undo_stack)
        self.notebook.add_tab(self.songbook, self.songbook_pane)
        self.songbook_pane.set_songbook(self.songbook)
        
      self.notebook.add_tab(self.song)
      
      self.partitions_pane = PartitionsPane(self, self.undo_stack)
      self.notebook.add_tab(self.song.partitions[0], self.partitions_pane, 0)
      self.playlist_pane = PlaylistPane(self, self.undo_stack)
      self.playlist_pane.set_playlist(self.song.playlist)
      self.notebook.add_tab(self.song.mesures[0], self.playlist_pane)
      self.notebook.add_tab(None)
      self.notebook.show_all()
      self.notebook.set_current_page(-1)
      
      
    self.scrolled = gtk.ScrolledWindow()
    self.scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    self.scrolled.set_size_request(800, 260)
    self.canvas = Canvas(self, self.song, self.scrolled)
    self.scrolled.add(self.canvas)
    
    if   globdef.config.SIDE_BAR_POSITION == "left":
      self.paned.pack1(self.notebook, 0, 1)
      self.paned.pack2(self.scrolled, 1, 1)
    elif globdef.config.SIDE_BAR_POSITION == "right":
      self.paned.pack2(self.notebook, 0, 1)
      self.paned.pack1(self.scrolled, 1, 1)
    elif globdef.config.SIDE_BAR_POSITION == "top":
      self.paned.pack1(self.notebook, 0, 1)
      self.paned.pack2(self.scrolled, 1, 1)
    elif globdef.config.SIDE_BAR_POSITION == "bottom":
      self.paned.pack2(self.notebook, 0, 1)
      self.paned.pack1(self.scrolled, 1, 1)
    elif (globdef.config.SIDE_BAR_POSITION == "floating_horizontal") or (globdef.config.SIDE_BAR_POSITION == "floating_vertical"):
      self.floating.add(self.notebook)
      self.floating.show_all()
      self.vbox.pack_end(self.scrolled, 1, 1)
    else:
      self.vbox.pack_end(self.scrolled, 1, 1)
      
    if self.paned:
      self.paned.set_border_width(3)
      self.vbox.pack_end(self.paned, 1, 1)
      
    self.show_all()
    
  def set_song(self, song):
    BaseApp.set_song(self, song)
    
    if getattr(self, "canvas", None):
      zoom = self.canvas.zoom
      self.canvas.destroy()
      self.canvas = songwrite2.canvas.Canvas(self, self.song, self.scrolled, zoom)
      self.scrolled.add(self.canvas)
      
    self.set_title(u"Songwrite2 -- %s" % song.title)
    if hasattr(self, "notebook"):
      self.show_all()

  def set_menu_enable(self, menu, enable): menu.set_sensitive(enable)
  def set_menu_label (self, menu, label ): menu.get_child().set_text(label)
    
  def prompt_open_filename(self, exts, format = u""):
    if exts == ".sw.xml": filter = sw_xml_filter
    else:
      filter = gtk.FileFilter()
      filter.set_name(format)
      if isinstance(exts, basestring): exts = [exts]
      for ext in exts:
        filter.add_pattern(u"*" + ext)
        
    dialog = gtk.FileChooserDialog(_(u"Open..."), self, gtk.FILE_CHOOSER_ACTION_OPEN, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))
    dialog.set_default_response(gtk.RESPONSE_OK)
    dialog.add_filter(filter)
    dialog.set_filename(os.path.join(globdef.config.LAST_DIR, u"DOES_NOT_EXIST"))
    response = dialog.run()
    filename = dialog.get_filename()
    dialog.destroy()
    if response != gtk.RESPONSE_OK: return
    return filename
  
  def prompt_save_filename(self, exts, format = u""):
    if exts == ".sw.xml": filter = sw_xml_filter
    else:
      filter = gtk.FileFilter()
      filter.set_name(format)
      if isinstance(exts, basestring): exts = [exts]
      for ext in exts:
        filter.add_pattern(u"*" + ext)
        
    dialog = gtk.FileChooserDialog(_(u"Save as..."), self, gtk.FILE_CHOOSER_ACTION_SAVE, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_SAVE, gtk.RESPONSE_OK))
    dialog.set_property("do-overwrite-confirmation", 1)
    dialog.set_default_response(gtk.RESPONSE_OK)
    
    dialog.add_filter(filter)
    dialog.set_current_folder(os.path.dirname(self.filename) or globdef.config.LAST_DIR)
    dialog.set_current_name(os.path.basename(self.filename).replace("sw.xml", exts[0]))
    response = dialog.run()
    filename = dialog.get_filename()
    dialog.destroy()
    if response != gtk.RESPONSE_OK: return
    return filename
  
  def on_open(self, event = None):
    if self.check_save(): return
    
    filename = self.prompt_open_filename(".sw.xml")
    if not filename: return
    self.open_filename(filename, do_check_save = 0) # check_save has already been done before asking the filename
    
  def check_save(self):
    if not self.song: return 0
    
    if self.undo_stack.undoables != self.last_undoables:
      # The undoable actions have changed => the song need to be saved.
      dialog = gtk.MessageDialog(self, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING, message_format = _(u"Save the modification before closing?"))
      dialog.add_buttons(
        _(u"Close without saving"), 0,
        gtk.STOCK_CANCEL, 1,
        gtk.STOCK_SAVE, 2,
        )
      dialog.set_default_response(1)
      response = dialog.run()
      dialog.destroy()
      if   response == 0: return 0
      if   response == 1: return 1
      elif response == 2:
        self.on_save()
        return self.check_save() # The user may have canceled the save as dialog box !
      
    else: return 0
    
          
    



      

class EditNotebook(gtk.Notebook):
  def __init__(self, undo_stack, tab_pos, *objs):
    gtk.Notebook.__init__(self)
    self.undo_stack      = undo_stack
    self.set_tab_pos(tab_pos)
    self.current         = 0
    self.tab_width       = 0
    self.objs            = []
    self.icon_panes      = []
    self.attribute_panes = []
    self.extra_widgets   = []
    for obj in objs: self.add_tab(obj)
    
    self.connect("switch-page", self.on_switch_page)
    
  def delete_tab(self, i):
    self.remove_page(i)
    del self.objs           [i]
    del self.icon_panes     [i]
    del self.attribute_panes[i]
    del self.extra_widgets  [i]
    
  def add_tab(self, obj = None, widget = None, widget_extend = 1):
    self.insert_tab(len(self.objs), obj, widget, widget_extend)
    
  def insert_tab(self, i, obj = None, widget = None, widget_extend = 1):
    label          = gtk.Label()
    icon_pane      = editor.IconPane     ("Gtk", self, 1, 1, 0)
    attribute_pane = editor.AttributePane("Gtk", self, None, self.undo_stack)
    self.objs           .insert(i, None)
    self.icon_panes     .insert(i, icon_pane)
    self.attribute_panes.insert(i, attribute_pane)
    self.extra_widgets  .insert(i, widget)
    
    scroll = gtk.ScrolledWindow()
    scroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
    scroll.add_with_viewport(attribute_pane)
    scroll.get_child().set_shadow_type(gtk.SHADOW_NONE)
    scroll.set_border_width(5)
    
    if widget:
      if self.get_tab_pos() in (gtk.POS_LEFT, gtk.POS_RIGHT): page = gtk.HBox()
      else:                                                   page = gtk.VBox()
      page.pack_start(scroll, 1, 1)
      if self.get_tab_pos() in (gtk.POS_LEFT, gtk.POS_RIGHT): page.pack_start(gtk.VSeparator(), 0, 0)
      else:                                                   page.pack_start(gtk.HSeparator(), 0, 0)
      page.pack_end  (widget, widget_extend, 1)
    else:
      page = scroll
      
    self.insert_page(page, icon_pane, i)
    self.set_tab_label_packing(page, 1, 0, gtk.PACK_START)
    
    if obj: self.edit(i, obj)
    
    if (i <= self.current) and (self.current < len(self.objs) - 1): self.current += 1
    
    self.show_all()
    
  def edit(self, i, obj, force = 0):
    if i < 0: i = len(self.objs) + i
    
    if not self.objs[i] is obj:
      self.objs[i] = obj
      self.icon_panes[i].edit(obj)
      if self.extra_widgets[i]: self.extra_widgets[i].edit(obj)
      
      self.icon_pane_changed(self.icon_panes[i])
      if (i == self.current) or force:
        self.attribute_panes[i].edit(obj)
      if (i == len(self.objs) - 2): self.check_mesures()
      
      
  def check_mesures(self):
    if isinstance(self.objs[-2], introsp.ObjectPack):
      numbers = [m.get_number() for m in self.objs[-2].objects]
      if numbers:
        label = _(u"Bars #%s to #%s") % (min(numbers) + 1, max(numbers) + 1)
        if self.current == len(self.objs) - 2: label = u"<b>%s</b>" % label
        self.icon_panes[-2].label.set_markup(label)
        
  def on_switch_page(self, notebook, page, i):
    self.icon_pane_changed(self.icon_panes[self.current])
    self.icon_panes[self.current].config(1, 1, 0)
    self.icon_panes[i           ].config(0, 1, 1)
    self.icon_pane_changed(self.icon_panes[self.current])
    old_current = self.current
    self.current = i
    if (self.current == len(self.objs) - 2) or (old_current == len(self.objs) - 2): self.check_mesures()
    self.attribute_panes[i].edit(self.objs[i])
    self.attribute_panes[i].queue_resize()
    self.icon_panes[i].show_all()
    self.icon_pane_changed(self.icon_panes[i])
    
  def icon_pane_changed(self, icon_pane):
    if self.get_tab_pos() in (gtk.POS_LEFT, gtk.POS_RIGHT):
      icon_pane.set_size_request(-1, -1)
      tab_width = icon_pane.size_request()[0]
      if tab_width > self.tab_width: self.tab_width = tab_width
      for icon_pane in self.icon_panes: icon_pane.set_size_request(self.tab_width, -1)
    else:
      try:
        icon_pane.label.set_property("visible", 0)
        icon_pane.image.set_alignment(0.5, 0.5)
      except: pass



class SongbookPane(gtk.HBox):
  def __init__(self, main, undo_stack = None):
    gtk.HBox.__init__(self, 0, 0)
    
    self.main           = main
    self.hierarchy_pane = editor.HierarchyPane("Gtk", self, self.on_song_selected, undo_stack)
    self.childhood_pane = editor.ChildhoodPane("Gtk", self,                        undo_stack)
    
    self.hierarchy_pane.set_childhood_pane(self.childhood_pane)
    
    scroll = gtk.ScrolledWindow()
    scroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
    scroll.set_shadow_type(gtk.SHADOW_IN)
    scroll.add(self.hierarchy_pane)
    self.set_border_width(5)
    
    self.pack_start(scroll, 1, 1)
    self.pack_start(self.childhood_pane, 0, 0)
    
  def on_song_selected(self, song_ref):
    if isinstance(song_ref, introsp.ObjectPack): song_ref = song_ref.objects[0]
    if isinstance(song_ref, model.SongRef):
      if self.main.check_save(): return
      self.main.set_song(song_ref.get_song())
      
      
  def set_songbook(self, songbook):
    self.hierarchy_pane.edit(songbook)
    
  def edit(self, o): pass
  
App.SongbookPane = SongbookPane


class PlaylistPane(gtk.HBox):
  def __init__(self, main, undo_stack = None):
    gtk.HBox.__init__(self, 0, 0)
    
    self.main           = main
    self.hierarchy_pane = editor.HierarchyPane("Gtk", self, lambda o: None, undo_stack)
    self.childhood_pane = editor.ChildhoodPane("Gtk", self,                 undo_stack)
    
    self.hierarchy_pane.set_childhood_pane(self.childhood_pane)
    
    scroll = gtk.ScrolledWindow()
    scroll.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
    scroll.set_shadow_type(gtk.SHADOW_IN)
    scroll.add(self.hierarchy_pane)
    self.set_border_width(5)
    
    self.pack_start(scroll, 1, 1)
    self.pack_start(self.childhood_pane, 0, 0)
    
  def set_playlist(self, playlist):
    self.hierarchy_pane.edit(playlist)
    
  def edit(self, o): pass
  


class PartitionsPane(gtk.HBox):
  def __init__(self, main, undo_stack = None):
    gtk.HBox.__init__(self)
    
    self.main           = main
    self.childhood_pane = editor.ChildhoodPane("Gtk", self, undo_stack)
    
    self.pack_start(self.childhood_pane, 0, 0)
    self.set_border_width(5)
    
  def edit(self, partition):
    self.childhood_pane.edit(partition.song, partition)
    

