# -*- coding: utf-8 -*-

# Songwrite 2
# Copyright (C) 2001-2007 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys, os, cStringIO as StringIO, tempfile, shutil
#import string as string_module # not a guitar string !

import songwrite2.globdef  as globdef
import songwrite2.model    as model
import songwrite2.plugins  as plugins
from   songwrite2.plugins.ps_native.ps_canvas import BaseCanvas, PSCanvas, guess_zoom

def escape_latex(s):
  return s.replace(u"°", u"\\symbol{6}").replace(u"#", u"\\#").replace(u"_", u"\\_").replace(u"~", u"\\~").replace(u"&", u"\\&").replace(u"%", u"\\%").replace(u"$", u"\\$").replace(u"^", u"\\^")

lang_iso2latex = { # Languages supported by LaTeX with Babel
  u"fr" : u"\\usepackage[francais]{babel}",
  u"en" : u"\\usepackage[american]{babel}",
  u"it" : u"\\usepackage[italian]{babel}",
  u"sv" : u"\\usepackage[swedish]{babel}",
  u"es" : u"\\usepackage[spanish]{babel}",
  u"pt" : u"\\usepackage[portuges]{babel}",
  u"de" : u"\\usepackage[german]{babel}",
  }

def latexify_songbook(tmp_dir, songbook):
  latexes = []
  lang    = None
  
  for song_ref in songbook.song_refs:
    song = song_ref.get_song()
    latexes.append(latexify_song(tmp_dir, song, 1))
    if not lang: lang = song.lang
    
  open(os.path.join(tmp_dir, "main.latex"), "w").write((u"""
\\documentclass[%s,10pt]{article}
\\usepackage[T1]{fontenc}
\\usepackage[utf8]{inputenc}
\\usepackage[lmargin=1.0cm,rmargin=1.0cm,tmargin=1.0cm,bmargin=2.0cm]{geometry}
\\usepackage{graphicx}
\\usepackage{ae,aecompl}
%s

\\begin{document}

\\sloppy

\\title {%s}
\\author{%s}
\\date  {}
\\maketitle
\\vfill
%s
\\tableofcontents
\\pagebreak

%s


\\end{document}
""" % (globdef.config.PAGE_FORMAT,
       lang_iso2latex[lang],
       escape_latex(songbook.title),
       escape_latex(songbook.authors),
       escape_latex(songbook.comments),
       u"""
\\pagebreak
""".join(latexes),
       )).encode("utf8"))


PSID = 0
def latexify_song(tmp_dir, song, songbook = 0):
  global PSID
  
  latex = StringIO.StringIO()
  
  # Skip hidden view
  #partitions = filter(lambda partition: partition.view.__class__.__name__ != "HiddenView", song.partitions)
  partitions = song.partitions
  
  # Use literal string r"..." because of the backslash !!!
  if songbook:
    latex.write(u"""
\\addcontentsline{toc}{subsection}{%s}
\\begin{center}\\begin{LARGE} %s \\end{LARGE}\\end{center}
\\begin{center}\\begin{large} %s \\end{large}\\end{center}
\\begin{center}\\begin{large} %s \\end{large}\\end{center}

""" % (escape_latex(song.title),
       escape_latex(song.title),
       escape_latex(song.authors),
       escape_latex(song.date()),
       ))
    
  else:
    latex.write(u"""
\\documentclass[%s,10pt]{article}
\\usepackage[T1]{fontenc}
\\usepackage[utf8]{inputenc}
\\usepackage[lmargin=1.0cm,rmargin=1.0cm,tmargin=1.0cm,bmargin=2.0cm]{geometry}
\\usepackage{graphicx}
\\usepackage{ae,aecompl}
%s

\\begin{document}

\\sloppy

\\title {%s}
\\author{%s}
\\date  {%s}
\\maketitle

""" % (globdef.config.PAGE_FORMAT,
       lang_iso2latex[song.lang],
       escape_latex(song.title),
       escape_latex(song.authors),
       escape_latex(song.date()),
       ))

  if song.comments:
    latex.write(u"""
%s

\\smallskip{}
""" % escape_latex(song.comments).replace(u"\\n", u"\\n\\n"))

  max_time = max([10] + [partition.end_time() for partition in song.partitions if isinstance(partition, model.Partition)])
  max_width = {
    "a3paper"    : 1468,
    "a4paper"    : 1080,
    "a5paper"    : 681,
    "letterpaper": 1010,
    }[globdef.config.PAGE_FORMAT]
  zoom = guess_zoom(song) * song.printsize / 20.0

  max_mesures_width = 0
  mesuress = []
  selected_mesures = []
  canvas = BaseCanvas(song)
  width = canvas.start_x
  for mesure in song.mesures:
    if mesure.time >= max_time: break
    mesure_width = mesure.duration * zoom
    if width + mesure_width > max_width:
      mesuress.append(selected_mesures)
      if width > max_mesures_width: max_mesures_width = width
      selected_mesures = []
      width = canvas.start_x
    width += mesure_width
    selected_mesures.append(mesure)
  if selected_mesures:
    if width > max_mesures_width: max_mesures_width = width
    mesuress.append(selected_mesures)
    
  zoom = zoom * (max_width - canvas.start_x) / (max_mesures_width - canvas.start_x)
  
  for mesures in mesuress:
    ps_filename = os.path.join(tmp_dir, "ps_file_%s.ps" % PSID)
    PSCanvas(song, ps_filename, mesures[0].time, mesures[-1].end_time(), zoom)
    latex.write(u"""
\\noindent \\begin{flushleft}
\\includegraphics[clip,scale=0.5]{ps_file_%s.ps}
\\par\\end{flushleft}
""" % PSID)
    PSID += 1

    if not mesures is mesuress[-1]:
      latex.write(u"""
\\smallskip{}
""")
  
  
  
  latex.write(u"""
\\vfill{}
""")
  
  for partition in partitions:
    if isinstance(partition, model.Lyrics):
      latex.write(u"""
\\begin{verse}
""")
      if partition.header: latex.write(u"""
\\subsection*{%s}
""" % escape_latex(partition.header))
      
      latex.write(escape_latex(partition.text.replace(u"_-\t", u"").replace(u"_\t", u" ").replace(u"-\t", u"").replace(u"\t", u" ").replace(u"_", u"").replace(u" ,", u",")))
      latex.write(u"""
\\end{verse}
""")
      

  if not songbook:
    latex.write(u"""
\\end{document}
""")

  if songbook:
    return latex.getvalue()
  else:
    open(os.path.join(tmp_dir, "main.latex"), "w").write(latex.getvalue().encode("utf8"))


PDF_PAGE_FORMATS = {
  "a3paper"     : "a3",
  "a4paper"     : "a4",
  "a5paper"     : "a5",
  "letterpaper" : "letter",
  }

def do(command):
  print >> sys.stderr
  print >> sys.stderr, "Running '%s'..." % command
  os.system(command)
  
def psify(song, pdf = 0):
  tmp_dir = tempfile.mkdtemp()
  print >> sys.stderr
  print >> sys.stderr, "Using temporary directory '%s'..." % tmp_dir
  
  if isinstance(song, model.Songbook):
    latexify_songbook(tmp_dir, song)
  else:
    latexify_song    (tmp_dir, song)
    
  do("cd %s; latex main.latex" % tmp_dir)
  do("cd %s; latex main.latex" % tmp_dir) # For table of content
  do("cd %s; dvips -o main.ps main.dvi" % tmp_dir)
  if pdf:
    do("cd %s; ps2pdf -sDEFAULTPAPERSIZE=%s main.ps" % (tmp_dir, PDF_PAGE_FORMATS[globdef.config.PAGE_FORMAT]))
    r = open(os.path.join(tmp_dir, "main.pdf")).read()
  else:
    r = open(os.path.join(tmp_dir, "main.ps")).read()

  shutil.rmtree(tmp_dir)
  return r
